#include "ptoc.h"

/***********************************************************/
/*                                                         */
/* ICS 4410                                                */
/* Fall '87                                                */
/*                                                         */
/* Lynn R. Akers, Jr.                                      */
/*                                                         */
/* This program has been modified to run on an IBM PC or   */
/* compatable using TURBO Pascal version 3.0.              */
/*                                                         */
/* It has only been tested for MICRO, so some bugs may     */
/* still exist.                                            */
/*                                                         */
/***********************************************************/

/* */
/*$R+*/


/*version of the SAM assembler for the MACC2 byte-addressable machine*/
/*        Author: C.T. Wilkes (from original by A.B. Maccabe)        */
/*                   Last modified: 84/11/26 CTW                     */
typedef set_template<16> word1;

typedef set_template<8> byte1;

const int MAXINT = 32768;

typedef int wordrange;
const int min_wordrange = -MAXINT ;  
const int max_wordrange = MAXINT - 1 ;

typedef signed char byterange;
const int min_byterange = -128;
const int max_byterange = 127;

enum opkind {oin, ia, is, im, ident, fn, fa, fs, fm, fd, bi, bo, ba, ic,
    fc, jsr, bkt, ld, sto, lda, flt, fix, j, sr, sl, rd, wr, trng, halt2,
    nop, clr, realdir, stringdir, intdir, skipdir, labeldir, last_opkind};

enum errorkind {noerr, unkopname, badregaddr, badgenaddr, badint,
             badreal, badstr, badname, illregaddr, illimedaddr, badshftamt,
             badstring, invalidname, last_errorkind};
enum warnkind {longname, missingr, missingnum, missingrp, missingcomma,
            namedefined, badskip, esctooshort, strtooshort, textfollows, last_warnkind};

//struct icrec {
typedef   union icrec{
      word1 wdf; // 4 bytes
      struct pad {
//      wordrange p1;
      wordrange sintf; //2 bytes
      }pad;
   }icrec;
//};

/*
struct bcrec {
   union {
      byte1 bytef; 
      byterange hintf; 
   };
};
*/
typedef union bcrec {
   byte1 bytef; //4 bytes
   struct pad {
    byterange p3, p2, p1; //1 byte each
   	byterange hintf; // 1 byte
  } pad;
} bcrec;

typedef array< 1 , 25 ,char> stringtype;



typedef array<1,7,char> name;

typedef struct symrec* symptr;
struct symrec {
   name id;
   symptr left, right;
   integer patch, loc;
};

word1 inop, iaop, isop, imop, idop, fnop, faop, fsop, fmop, fdop,
biop, boop, baop, icop, fcop, jsrop, bktop, ldop, stoop, ldaop,
fltop, fixop, jop, srop, slop, rdop, wrop, trngop, haltop;

text infile;

boolean saved;   /*flag for one character lookahead*/
char ch;    /*Current character from the input*/

array<0,MAXINT,byte1> mem;      /*Memory Image being created*/ 
unsigned lc;    /*Location Counter*/
file<byterange> memf;       /*File for the memory image*/

symptr symbols;
integer line;   /*Number of the current input line*/
boolean listing;

errorkind error;
boolean errs;
boolean warning;
boolean morewarn;
array<1,10,warnkind> warns;
integer windex;

icrec icr; /*Integer/Word Conversion Record*/
bcrec bcr; /*Halfint/Byte Conversion Record*/
integer i; /*loop index*/

stringtype arg, memfname, srcfname;



    void wordtobytes (word1 w,  byte1& hibyte, byte1& lobyte)
    {
        integer i;

    lobyte = byte1::of(eos);
    hibyte = byte1::of(eos);
    for( i = 0; i <= 7; i ++)
       {
       if (w.has(i)) 
          lobyte = lobyte + byte1::of(i, eos);
       if (w.has(i+8)) 
           hibyte = hibyte + byte1::of(i, eos);
       }   /*for*/
    }    /*procedure wordtobytes*/


    void bytestoword (byte1 hibyte, byte1 lobyte,  word1& w)
    {
        integer i;

    w = word1::of(eos);
    for( i = 0; i <= 7; i ++)
       {
       if (lobyte.has(i)) 
          w = w + word1::of(i, eos);
       if (hibyte.has(i)) 
          w = w + word1::of(i+8, eos);
       }   /*for*/
    }    /*procedure bytestoword*/

void showbyte(int i)
{	bcrec bcr;
           bcr.bytef = mem[i];
           output << bcr.pad.hintf << "\n";

}

    void insertmem (word1 w)
    {
        byte1 lobyte, hibyte;

    wordtobytes (w, hibyte, lobyte);
    mem[lc] = hibyte;
    mem[lc+1] = lobyte;
    lc = lc + 2;
    }    /*procedure insertmem*/

    void checktab(symptr cur)
    {
    if (cur != nil) 
       {
       checktab(cur->left);
       if (cur->loc < 0) 
          {
          warning = true;
          output << "WARNING -- " << cur->id << " Undefined" << NL;
          }
       checktab(cur->right);
       }   /*if*/
    }    /*procedure checktab*/


    void warn (warnkind w)
    {
    if (! morewarn) 
       windex = 1;
    morewarn = true;
    warns[windex] = w;
    windex = windex + 1;
    }    /*procedure warn*/


    void printwarn()
    {
        integer i;

    output << "WARNING -- ";
    switch (warns[1]) {
     case textfollows: output << "Text follows instruction"; break;
     case esctooshort: output << "Need 3 digits to specify an unprintable character"; break;
     case strtooshort: output << "Missing \" in string"; break;
     case badskip: output << "Skip value must be positive, Skip directive ignored"; break;
     case namedefined: output << "Name already defined, earlier definition lost"; break;
     case longname: output << "Name too long, only 7 characters used"; break;
     case missingr: output << "Missing R in Register Address"; break;
     case missingnum: output << "Missing Number in Register Address (0 assumed)"; break;
     case missingrp: output << "Missing \")\" in Indexed Address"; break;
     case missingcomma: output << "Missing \",\"";
     break;
 }    /*case*/
 output << " on line " << format(line, 1) << NL;

 for( i = 2; i <= windex - 1; i ++)
     warns[i - 1] = warns[i];
 windex = windex - 1;
 if (windex <= 1) 
     morewarn = false;
    }    /*procedure printwarn*/


    void inregaddr (word1& w, integer reg, integer hbit)
    /*Insert a register address into a word*/
    {
    if (reg > 7)  {
        w = w + word1::of(hbit, eos);
        reg = reg - 8;
    }    /*if reg > 7*/
    hbit = hbit - 1;
    if (reg > 3)  {
       w = w + word1::of(hbit, eos);
       reg = reg - 4;
       }    /*if reg > 3*/
    hbit = hbit - 1;
    if (reg > 1)  {
       w = w + word1::of(hbit, eos);
       reg = reg - 2;
       }    /*if reg > 1*/
    hbit = hbit - 1;
    if (reg > 0) 
       w = w + word1::of(hbit, eos);
    }    /*procedure inregaddr*/


    void getch()
    /*Get a character from the input -- character may have been saved*/
    {
 if (! saved) 
     if (eof ( infile )) 
        ch = '%';
     else if (! eoln ( infile )) 
        {
        do {
          infile >> ch;
          if (listing) 
            output << ch;
        } while (!(((ch != ' ') && (ch != chr(9))) || eoln ( infile )));
        if (ch == '%')  {
           while (! eoln ( infile )) 
              {
              infile >> ch;
              if (listing) 
              output << ch;
              }    /*while*/
           ch = '%';
           }
       }
          else
          ch = '%';
   else
     saved = false;
    }     /*procedure getch*/


    void scanname(name& id)
    {
       integer i;

    for( i = 1; i <= 7; i ++)
       id[i] = ' ';
    i = 1;
    while ((set::of(range('a','z'), range('A','Z'), range('0','9'),'$', eos).has(ch)) && (i <= 7)) 
       {
       id[i] = ch;
       i = i + 1;
       getch();
       }    /*while*/
    if (set::of(range('a','z'), range('A','Z'), range('0','9'),'$', eos).has(ch)) 
       warn(longname);
    while (set::of(range('a','z'), range('A','Z'), range('0','9'),'$', eos).has(ch)) 
       getch();
    saved = true;
    }    /*procedure scanname*/


    symptr findname(name id)
    {
       symptr temp;
       boolean found;

    symptr findname_result;
    temp = symbols;
    found = false;
    while (! found && (temp != nil)) 
       if (temp->id == id) 
          found = true;
       else if (temp->id > id) 
          temp = temp->left;
       else
          temp = temp->right;
    findname_result = temp;
    return findname_result;
    }    /*function findname*/


    symptr inname(name id)
    {
       symptr cur, prev;

    symptr inname_result;
    cur = symbols;
    prev = nil;
    while (cur != nil) 
       {
       prev = cur;
       if (cur->id > id) 
          cur = cur->left;
       else
          cur = cur->right;
       }    /*while*/
    cur = new symrec;
    cur->left = nil;
    cur->right = nil;
    cur->id = id;
    if (prev == nil) 
       symbols = cur;
    else if (prev->id > id) 
       prev->left = cur;
    else
       prev->right = cur;
    inname_result = cur;
    return inname_result;
    }    /*function inname*/



struct ccrec {
  union {
    word1 wf;  
    struct { char c4,c3,c1, c2; } s_false;//was Char c2,c1
//    struct { char c2,c1,c3, c4; } s_false;//was Char c2,c1
  };                        /*Inverted Bytes!!!*/
};


    void scanstr()
    {
       boolean one;
       ccrec ccr;
       integer ival;

    one = true;
    getch();
    if (ch == '"') 
       {
       if (! eoln ( infile )) 
          {
          infile >> ch;
          if (listing) 
             output << ch;
          }    /*if*/
       while ((ch != '"') && ! eoln ( infile )) 
          {
          if (ch == ':') 
             {
             if (! eoln ( infile )) 
                {
                infile >> ch;
                if (listing) 
                   output << ch;
                if (set::of(range('0','9'), eos).has(ch)) 
                   {
                   ival = ord(ch) - ord('0');
                   if (! eoln ( infile )) 
                      {
                      infile >> ch;
                      if (listing) 
                      output << ch;
                      if (set::of(range('0','9'), eos).has(ch)) 
                         {
                         ival = ival * 10 + ord(ch) - ord('0');
                         if (! eoln ( infile )) 
                            {
                            infile >> ch;
                            if (listing) 
                               output << ch;
                            if (set::of(range('0','9'), eos).has(ch)) 
                               ival = ival * 10 + ord(ch) - ord('0');
                            else
                               warn(esctooshort);
                            }
                         else
                            warn(esctooshort);
                         }
                      else
                         warn(esctooshort);
                      }
                   else
                      warn(esctooshort);
                   ch = chr(ival);
                   }
                }
             else
                warn(esctooshort);
             }
             if (one) 
                {
                one = false;
                ccr.s_false.c1 = ch;
                }
             else
                {
                one = true;
                ccr.s_false.c2 = ch;
                insertmem (ccr.wf);
                }    /*else*/
             if (! eoln ( infile )) 
                {
                infile >> ch;
                if (listing) 
                   output << ch;
                }   /*if*/
          }    /*while*/
       if (one) {
          ccr.s_false.c1 = chr(0);
          ccr.s_false.c2 = chr(0); //CLZ
        }
       else
         ccr.s_false.c2 = chr(0);
       insertmem (ccr.wf);
       if (ch != '"') 
          warn(strtooshort);
       }
    else
      error = badstring;
    }    /*procedure scanstr*/



//typedef set bytetype;

struct mrec {
   byte1 byte1a;
   byte1 byte2;
   byte1 byte3;
   byte1 byte4;
   byte1 byte5;
   byte1 byte6;
};

struct rcrec {
   union {
      real rf;  
      mrec mf; 
   };
};

    void scanreal(word1& w1, word1& w2)

    {
       rcrec rcr;
       real dval;


    rcr.rf = 0.0;
    getch();
    while (set::of(range('0','9'), eos).has(ch)) 
       {
       rcr.rf = rcr.rf * 10 + ord(ch) - ord('0');
       getch();
       }    /*while*/
    if (ch == '.') 
       {
       getch();
       dval = 10.0;
       while (set::of(range('0','9'), eos).has(ch)) 
          {
          rcr.rf = rcr.rf + (ord(ch) - ord('0')) / dval;
          dval = dval * 10.0;
          getch();
          }   /*while*/
       }
    else
       saved = true;
    bytestoword ( rcr.mf.byte1a, rcr.mf.byte4, w1 );
    bytestoword ( rcr.mf.byte5, rcr.mf.byte6, w2 );
    }    /*procedure scanreal*/


    void scanint(word1& w)
    /*Get an integer from the input stream*/
    {
       integer temp;
       boolean neg;

    neg = false;
    temp = 0;
    getch();
    if (set::of('-', '+', eos).has(ch)) 
       {
       if (ch == '-') 
       neg = true;
       getch();
       }    /*if Ch in ['-', '+']*/
    while (set::of(range('0','9'), eos).has(ch)) 
       {
       temp = temp * 10 + ord(ch) - ord('0');
       getch();
       }    /*while*/
    saved = true;     /*Note the lookahead*/
    if (neg) 
       temp = -temp;
    if ((temp > MAXINT) || (temp < -MAXINT)) 
       error = badint;
    else {
       icr.pad.sintf = temp;
       w = icr.wdf;
    }
    }    /* scanint */


    integer getregaddr()
    /*Get a register address from the input stream*/
    {
       integer temp;


    integer getregaddr_result;
    getch();
    if (ch == 'R') 
       getch();
    else
       warn(missingr);
    if (set::of(range('0','9'), eos).has(ch)) 
       {      /*Check first digit*/
       temp = ord(ch) - ord('0');
       getch();
       if (set::of(range('0','9'), eos).has(ch))  		/*Check for two digits*/
          temp = temp * 10 + ord(ch) - ord('0');
       else
          saved = true;
       if (temp > 15) 
          error = badregaddr;
       }
    else  /*if Ch is a digit*/
       warn(missingnum);

 getregaddr_result = temp;
    return getregaddr_result;
    }    /*function getregaddr*/


    void getgenaddr (opkind op, word1& w1, word1& w2, boolean& flag)
    {
        integer reg;
        name id;
        symptr idrec;


    flag = false;
    getch();
    if (ch == '*') 
       {
       w1 = w1 + word1::of(6, eos);
       getch();
       }    /*if Ch = '*'*/

    if (set::of(range('a','z'), range('A','Q'), range('S','Z'),'$', eos).has(ch)) 
       {
       flag = true;
       scanname(id);
       idrec = findname(id);
       if (idrec == nil) 
          {
          idrec = inname(id);
          idrec->loc = -1;
          idrec->patch = lc + 2;
          icr.pad.sintf = -1;
          }
       else if (idrec->loc == -1) 
          {
          icr.pad.sintf = idrec->patch;
          idrec->patch = lc + 2;
          }
       else
          icr.pad.sintf = idrec->loc;
       w2 = icr.wdf;
       getch();
       if (ch == '(') 
          {
          w1 = w1 + word1::of(5, eos);
          reg = getregaddr();
          if (error == noerr) 
             {
             inregaddr(w1, reg, 3);
             getch();
             if (ch != ')') 
                warn(missingrp);
             }   /*if Error in register address*/
          }
        else     /*Ch <> '('*/
           w1 = w1 + word1::of(4, eos);
       }   /* Letter other than 'R' */
    else if (set::of(range('0','9'), eos).has(ch)) 
       {
       saved = true;
       w1 = w1 + word1::of(4, eos);
       flag = true;
       scanint(w2);
       }
    else if (set::of('R', '#', '-', '+', '&', eos).has(ch)) 
       switch (ch) {
          case 'R' :  {   /*direct register*/
                 flag = false;
                 saved = true;
                 reg = getregaddr();
                 inregaddr(w1, reg, 3);
                 if ((set_of_enum(opkind)::of(jsr, bkt, lda, j, eos).has(op)) && ! (w1.has(6))) 
                    error = illregaddr;
                 }
                 break;
          case '#' : {   /*immediate*/
                flag = true;
                if (w1.has(6)) 
                   error = badgenaddr;
                else if (set_of_enum(opkind)::of(fn, fa, fs, fm, fd, fc, fix,
                               jsr, bkt, sto, j, rd, trng, eos).has(op)) 
                   error = illimedaddr;
                else if (w1 == wrop + word1::of(7, eos)) 
                   error = illimedaddr;
                else if (w1 == wrop + word1::of(10, 7, eos)) 
                   error = illimedaddr;
                else
                   {
                   w1 = w1 + word1::of(4, 5, eos);
                   scanint(w2);
                   }   /*else*/
                }
                break;
     case '-': case '+' : {   /*indexed*/
                w1 = w1 + word1::of(5, eos);
                flag = true;
                if (ch == '-') 
                   saved = true;
                scanint(w2);
                getch();
                if (ch == '(') 
                   {
                   reg = getregaddr();
                   if (error == noerr) 
                      {
                      inregaddr(w1, reg, 3);
                      getch();
                      if (ch != ')') 
                         warn(missingrp);
                      }   /*if Error in register address*/
                   }
                else   /*Ch <> '('*/
                   error = badgenaddr;
                }
                break;
          case '&' : {
                flag = true;
                if (w1.has(6)) 
                   error = badgenaddr;
                else
                   {
                   w1 = w1 + word1::of(4, 5, 6, eos);
                   scanint(w2);
                   }
                }
                break;
     }   /*case of Ch*/
 else
     error = badgenaddr;
    }    /*procedure getgenaddr*/


    void getbop(opkind& op, word1& wd)
    /*Get an operator or directive name that begins with 'B'*/
    {
    getch();
 if (set::of('A', 'I', 'K', 'O', eos).has(ch)) 
     switch (ch) {
  case 'A':
      {
   op = ba;
   wd = baop;
      }
      break;
  case 'I':
      {
   op = bi;
   wd = biop;
      }
      break;
  case 'K':
      {
   getch();
   if (ch == 'T')  {
       op = bkt;
       wd = bktop;
   } else     /*Ch <> 'T'*/
       error = unkopname;
      }
      break;
  case 'O':
      {
   op = bo;
   wd = boop;
      }
      break;
     }   /*case of Ch*/
 else      /*character does not legally follow `B'*/
     error = unkopname;
    }    /*procedure getBop*/


    void getfop(opkind& op, word1& wd)
    /*Get an operator or directive name that begins with 'F'*/
    {
        getch();
        if (set::of('A', 'C', 'D', 'I', 'L', 'M', 'N', 'S', eos).has(ch))  
            switch (ch) {
                case 'A':
                    {
                        op = fa;
                        wd = faop;
                    }
                    break;
                case 'C':
                    {
                        op = fc;
                        wd = fcop;
                    }
                    break;
                case 'D':
                    {
                        op = fd;
                        wd = fdop;
                    }
                    break;
                case 'I':
                    {
                        getch();
                        if (ch == 'X')  {
                            op = fix;
                            wd = fixop;
                        } else 		        /*Ch <> 'X'*/
                            error = unkopname;
                    }
                    break;
                case 'L':
                    {
                        getch();
                        if (ch == 'T')  {
                            op = flt;
                            wd = fltop;
                        } else 		        /*Ch <> 'T'*/
                            error = unkopname;
                    }
                    break;
                case 'M':
                    {
                        op = fm;
                        wd = fmop;
                    }
                    break;
                case 'N':
                    {
                        op = fn;
                        wd = fnop;
                    }
                    break;
                case 'S':
                    {
                        op = fs;
                        wd = fsop;
                    }
                    break;
            }   /*case of Ch*/
        else 					/*character does not legally follow `F'*/
            error = unkopname;
    }    /*procedure getFop*/


    void getiop(opkind& op, word1& wd)
    /*Get an operator or directive name that begins with 'I'*/
    {
        getch();
        if (set::of('A', 'C', 'D', 'M', 'N', 'S', eos).has(ch))  
            switch (ch) {
                case 'A':
                    {
                        op = ia;
                        wd = iaop;
                    }
                    break;
                case 'C':
                    {
                        op = ic;
                        wd = icop;
                    }
                    break;
                case 'D':
                    {
                        op = ident;
                        wd = idop;
                    }
                    break;
                case 'M':
                    {
                        op = im;
                        wd = imop;
                    }
                    break;
                case 'N':
                    {
                        getch();
                        if (ch == 'T')  
                            op = intdir;
                        else {
                            op = oin;
                            wd = inop;
                            saved = true;
                        }
                    }
                    break;
                case 'S':
                    {
                        op = is;
                        wd = isop;
                    }
                    break;
            }   /*case of Ch*/
        else 					/*character does not legally follow `I'*/
            error = unkopname;
    }    /*getIop*/


    void getjop(opkind& op, word1& wd)
    /*Get an operator or directive name that begins with 'J'*/
    {
        getch();
        if (set::of('E', 'G', 'L', 'M', 'N', 'S', eos).has(ch))  {
            op = j;  /*most are simple jumps--except JSR!!*/
            switch (ch) {
                case 'E':
                    {
                        getch();
                        if (ch == 'Q') 
                            wd = jop + word1::of(7, 8, eos);
                        else 
                            error = unkopname;
                    }
                    break;
                case 'G':
                    {
                        getch();
                        if (ch == 'E') 
                            wd = jop + word1::of(7, 9, eos);
                        else if (ch == 'T') 
                            wd = jop + word1::of(8, 9, eos);
                        else 
                            error = unkopname;
                    }
                    break;
                case 'L':
                    {
                        getch();
                        if (ch == 'E')  
                            wd = jop + word1::of(8, eos);
                        else if (ch == 'T') 
                            wd = jop + word1::of(7, eos);
                        else
                            error = unkopname;
                    }
                    break;
                case 'M':
                    {
                        getch();
                        if (ch == 'P')  
                            wd = jop;
                        else
                            error = unkopname;
                    }
                    break;
                case 'N':
                    {
                        getch();
                        if (ch == 'E') 
                            wd = jop + word1::of(9, eos);
                        else 
                            error = unkopname;
                    }
                    break;
                case 'S':
                    {
                        getch();
                        if (ch == 'R')  {
                            op = jsr;
                            wd = jsrop;
                        } else 		        /*Ch <> 'R'*/
                            error = unkopname;
                    }
                    break;
            }   /*case of Ch*/
        } else 				        /*Ch not in ['E','G',...]*/
            error = unkopname;
    }    /*procedure getJop*/


    void getlop(opkind& op, word1& wd)
    /*Get an operator or directive name that begins with 'L'*/
    {
        getch();
        if (set::of('A', 'D', eos).has(ch)) 
            switch (ch) {
                case 'A':
                    {
                        getch();
                        if (ch == 'B')  {
                            getch();
                            if (ch == 'E')  {
                                getch();
                                if (ch == 'L') 
                                    op = labeldir;
                                else
                                    error = unkopname;
                            } else
                                error = unkopname;
                        } else
                            error = unkopname;
                    }
                    break;
                case 'D':
                    {
                        getch();
                        if (ch == 'A')  {
                            op = lda;
                            wd = ldaop;
                        } else {
                            op = ld;
                            wd = ldop;
                            saved = true;
                        }
                    }
                    break;
            }   /*case Ch of*/
        else
            error = unkopname;
    }    /*procedure getLop*/


    void getrop(opkind& op, word1& wd)
    /*Get an operator or directive name that begins with 'R'*/
    {
        getch();
        if (ch == 'D')  {
            op = rd;
            getch();
            if (set::of('B', 'C', 'F', 'H', 'I', 'N', 'O', 'S', eos).has(ch)) 
                switch (ch) {
                    case 'B':
                        {
                            getch();
                            if (ch == 'D') 
                                wd = rdop + word1::of(8, eos);
                            else if (ch == 'W')  
                                wd = rdop + word1::of(7, 8, eos);
                            else
                                error = unkopname;
                        }
                        break;
                    case 'C':
                        {
                            getch();
                            if (ch == 'H') 
                                wd = rdop + word1::of(10, eos);
                            else 
                                error = unkopname;
                        }
                        break;
                    case 'F':
                        wd = rdop + word1::of(7, eos);
                        break;
                    case 'H':
                        {
                            getch();
                            if (ch == 'D') 
                                wd = rdop + word1::of(8, 9, eos);
                            else if (ch == 'W')  
                                wd = rdop + word1::of(7, 8, 9, eos);
                            else
                                error = unkopname;
                        }
                        break;
                    case 'I':
                        wd = rdop;
                        break;
                    case 'N':
                        {
                            getch();
                            if (ch == 'L') 
                                wd = rdop + word1::of(7, 8, 10, eos);
                            else
                                error = unkopname;
                        }
                        break;
                    case 'O':
                        {
                            getch();
                            if (ch == 'D') 
                                wd = rdop + word1::of(9, eos);
                            else if (ch == 'W') 
                                wd = rdop + word1::of(7, 9, eos);
                            else
                                error = unkopname;
                        }
                        break;
                    case 'S':
                        {
                            getch();
                            if (ch == 'T') 
                                wd = rdop + word1::of(7, 10, eos);
                            else
                                error = unkopname;
                        }
                        break;
                }   /*case of Ch*/
            else 				/*Ch not in ['B','C',...]*/
                error = unkopname;
        } else if (ch == 'E')  {
            getch();
            if (ch == 'A')  {
                getch();
                if (ch == 'L')  
                    op = realdir;
                else
                    error = unkopname;
            } else 				/*Ch <> 'A'*/
                error = unkopname;
        } else 				        /*Ch <> 'E'*/
            error = unkopname;
    }    /*procedure getRop*/


    void getsop(opkind& op, word1& wd)
    /*Get an operator or directive name that begins with 'S'*/
    {
        getch();
        if (set::of('K', 'L', 'R', 'T', eos).has(ch))  
            switch (ch) {
                case 'K':
                    {
                        getch();
                        if (ch == 'I')  {
                            getch();
                            if (ch == 'P') 
                                op = skipdir;
                            else
                                error = unkopname;
                        } else 		        /*Ch <> 'I'*/
                            error = unkopname;
                    }
                    break;
                case 'L':
                    {
                        getch();
                        if (set::of('C', 'E', 'O', 'Z', eos).has(ch))  {
                            op = sl;
                            switch (ch) {
                                case 'C':
                                    {
                                        getch();
                                        if (set::of('C', 'E', 'O', 'Z', eos).has(ch))  
                                            switch (ch) {
                                                case 'C':
                                                    wd = slop + word1::of(4, 5, 6, eos);
                                                    break;
                                                case 'E':
                                                    wd = slop + word1::of(5, 6, eos);
                                                    break;
                                                case 'O':
                                                    wd = slop + word1::of(4, 6, eos);
                                                    break;
                                                case 'Z':
                                                    wd = slop + word1::of(6, eos);
                                                    break;
                                            }   /*case of Ch*/
                                        else {
                                            saved = true;
                                            wd = slop + word1::of(4, 5, eos);
                                        }   /*else*/
                                    }
                                    break;
                                case 'E':
                                    wd = slop + word1::of(5, eos);
                                    break;
                                case 'O':
                                    wd = slop + word1::of(4, eos);
                                    break;
                                case 'Z':
                                    wd = slop;
                                    break;
                            }   /*case Ch of*/
                        } else 
                            error = unkopname;
                    }
                    break;
                case 'R':
                    {
                        getch();
                        if (set::of('C', 'E', 'O', 'Z', eos).has(ch))  {
                            op = sr;
                            switch (ch) {
                                case 'C':
                                    {
                                        getch();
                                        if (set::of('C', 'E', 'O', 'Z', eos).has(ch)) 
                                            switch (ch) {
                                                case 'C':
                                                    wd = srop + word1::of(4, 5, 6, eos);
                                                    break;
                                                case 'E':
                                                    wd = srop + word1::of(5, 6, eos);
                                                    break;
                                                case 'O':
                                                    wd = srop + word1::of(4, 6, eos);
                                                    break;
                                                case 'Z':
                                                    wd = srop + word1::of(6, eos);
                                                    break;
                                            }   /*case of Ch*/
                                        else {
                                            saved = true;
                                            wd = srop + word1::of(4, 5, eos);
                                        }   /*else*/
                                    }
                                    break;
                                case 'E':
                                    wd = srop + word1::of(5, eos);
                                    break;
                                case 'O':
                                    wd = srop + word1::of(4, eos);
                                    break;
                                case 'Z':
                                    wd = srop;
                                    break;
                            }   /*case Ch of*/
                        } else
                            error = unkopname;
                    }
                    break;
                case 'T':
                    {
                        getch();
                        if (ch == 'O')  {
                            op = sto;
                            wd = stoop;
                        } else if (ch == 'R')  {
                            getch();
                            if (ch == 'I')  {
                                getch();
                                if (ch == 'N')  {
                                    getch();
                                    if (ch == 'G')  
                                        op = stringdir;
                                    else
                                        error = unkopname;
                                } else 	        /*Ch <> 'N'*/
                                    error = unkopname;
                            } else 		/*Ch <> 'I'*/
                                error = unkopname;
                        } else 		        /*Ch <> 'R'*/
                            error = unkopname;
                    }
                    break;
            }   /*case Ch of*/
        else 
            error = unkopname;
    }    /*procedure getSop*/


    void getwop(opkind& op, word1& wd)
    /*Get an operator or directive name that begins with 'W'*/
    {
        getch();
        if (ch == 'R')  {
            op = wr;
            getch();
            if (set::of('B', 'C', 'F', 'H', 'I', 'N', 'O', 'S', eos).has(ch)) 
                switch (ch) {
                    case 'B':
                        {
                            getch();
                            if (ch == 'D') 
                                wd = wrop + word1::of(8, eos);
                            else if (ch == 'W') 
                                wd = wrop + word1::of(7, 8, eos);
                            else 
                                error = unkopname;
                        }
                        break;
                    case 'C':
                        {
                            getch();
                            if (ch == 'H')  
                                wd = wrop + word1::of(10, eos);
                            else
                                error = unkopname;
                        }
                        break;
                    case 'F':
                        wd = wrop + word1::of(7, eos);
                        break;
                    case 'H':
                        {
                            getch();
                            if (ch == 'D') 
                                wd = wrop + word1::of(8, 9, eos);
                            else if (ch == 'W') 
                                wd = wrop + word1::of(7, 8, 9, eos);
                            else 
                                error = unkopname;
                        }
                        break;
                    case 'I':
                        wd = wrop;
                        break;
                    case 'N':
                        {
                            getch();
                            if (ch == 'L') 
                                wd = wrop + word1::of(7, 8, 10, eos);
                            else
                                error = unkopname;
                        }
                        break;
                    case 'O':
                        {
                            getch();
                            if (ch == 'D') 
                                wd = wrop + word1::of(9, eos);
                            else if (ch == 'W') 
                                wd = wrop + word1::of(7, 9, eos);
                            else
                                error = unkopname;
                        }
                        break;
                    case 'S':
                        {
                            getch();
                            if (ch == 'T')  
                                wd = wrop + word1::of(7, 10, eos);
                            else 
                                error = unkopname;
                        }
                        break;
                }   /*case of Ch*/
            else 
                error = unkopname;
        } else 				        /*Ch <> 'R'*/
            error = unkopname;
    }    /*procedure getWop*/


    void proline()
    /*process the current line of input*/
    {
        word1 wd, wd2;
        byte1 b1, b2;
        opkind op;
        integer reg;
        boolean twowds;
        integer i1, i2;
        name id;
        symptr idrec;

        twowds = false;
        error = noerr;
        if (set::of('B', 'C', 'F', 'H', 'I', 'J', 'L',
                  'N', 'R', 'S', 'T', 'W', eos).has(ch)) 
            switch (ch) {
                case 'B':
                    getbop(op, wd);
                    break;
                case 'C':
                    {
                        getch();
                        if (ch == 'L')  {
                            getch();
                            if (ch == 'R')  {
                                op = clr;
                                wd = srop;
                            } else 		/*Ch <> 'R'*/
                                error = unkopname;
                        } else 		        /*Ch <> 'L'*/
                            error = unkopname;
                    }
                    break;
                case 'F':
                    getfop(op, wd);
                    break;
                case 'H':
                    {
                        getch();
                        if (ch == 'A')  {
                            getch();
                            if (ch == 'L')  {
                                getch();
                                if (ch == 'T')  {
                                    op = halt2;
                                    wd = haltop;
                                } else 	        /*Ch <> 'T'*/
                                    error = unkopname;
                            } else 		/*Ch <> 'L'*/
                                error = unkopname;
                        } else 		        /*Ch <> 'A'*/
                            error = unkopname;
                    }
                    break;
                case 'I':
                    getiop(op, wd);
                    break;
                case 'J':
                    getjop(op, wd);
                    break;
                case 'L':
                    getlop(op, wd);
                    break;
                case 'N':
                    {
                        getch();
                        if (ch == 'O')  {
                            getch();
                            if (ch == 'P')  {
                                op = nop;
                                wd = jop + word1::of(7, 8, 9, eos);
                            } else 		/*Ch <> 'P'*/
                                error = unkopname;
                        } else 		        /*Ch <> 'O'*/
                            error = unkopname;
                    }
                    break;
                case 'R':
                    getrop(op, wd);
                    break;
                case 'S':
                    getsop(op, wd);
                    break;
                case 'T':
                    {
                        getch();
                        if (ch == 'R')  {
                            getch();
                            if (ch == 'N')  {
                                getch();
                                if (ch == 'G')  {
                                    op = trng;
                                    wd = trngop;
                                } else	        /*Ch <> 'G'*/
                                    error = unkopname;
                            } else		/*Ch <> 'N'*/
                                error = unkopname;
                        } else		        /*Ch <> 'R'*/
                            error = unkopname;
                    }
                    break;
                case 'W':
                    getwop(op, wd);
                    break;
            }   /*case of Ch*/
        else 					/*Ch not a valid first letter*/
            error = unkopname;
        if (error == noerr)  {
            switch (op) {
                case clr:
                    {			        /*need to find a reg address*/
                        reg = getregaddr();
                        if (error == noerr)  {
                            inregaddr(wd, reg, 10);
                            insertmem (wd);
                        }   /*if Error = noerr*/
                    }
                    break;			/*SL, SR case*/
                case sl: case sr:
                    {			        /*need to find a reg address*/
                        reg = getregaddr();
                        if (error == noerr) 
                            inregaddr(wd, reg, 10);
                        getch();
                        if (ch != ',')  {
                            warn(missingcomma);
                            saved = true;
                        }
                        scanint(icr.wdf);
                        if (error == noerr)  {
                            reg = icr.pad.sintf;
                            if (reg == 16) 
                                reg = 0;
                            if ((reg < 16) && (reg >= 0))  {
                                inregaddr(wd, reg, 3);
                                insertmem (wd);
                            } else
                                error = badshftamt;
                        }   /*if Error = noerr*/
                    }
                    break;			/*SL, SR case*/
                case oin: case ia: case is: case im: case ident: case fn: case fa:
                case fs: case fm: case fd: case bi: case bo: case ba: case ic:
                case fc: case jsr: case bkt: case ld: case sto: case lda: case flt:
                case fix: case trng:
                    {
                        reg = getregaddr();
                        if (error == noerr)  {
                            inregaddr(wd, reg, 10);
                            getch();
                            if (ch != ',')  {
                                warn(missingcomma);
                                saved = true;
                            }    /*if Ch <> ','*/

                            getgenaddr(op, wd, wd2, twowds);

                            if (error == noerr) 
                                if (twowds)  {
                                    insertmem (wd);
                                    insertmem (wd2);
                                } else
                                    insertmem (wd);
                        }   /*if Error = noerr*/
                    }
                    break;			/*two-address case*/
                case j:
                    {
                        getgenaddr(op, wd, wd2, twowds);
                        if (error == noerr) 
                            if (twowds)  {
                                insertmem (wd); //output<<"wd\n"; 
                                insertmem (wd2); //output << "wd2\n";
                            } else
                                insertmem (wd);
                    }
                    break;			/*one general address case*/
                case rd: case wr:
                    {
                        twowds = false;
                        if (! ((wd.has(10)) && (wd.has(8)))) 
                            getgenaddr(op, wd, wd2, twowds);
                        if (error == noerr) 
                            if (twowds)  {
                                insertmem (wd);
                                insertmem (wd2);
                            } else
                                insertmem (wd);
                    }
                    break;			/*one general address case*/
                case nop: case halt2:
                    insertmem (wd);
                    break;
                case intdir:
                    {
                        scanint(wd);
                        wordtobytes (wd, mem[lc], mem[lc+1]);
                        lc = lc + 2;
                    }
                    break;
                case realdir:
                    {
                        scanreal(wd, wd2);
                        wordtobytes (wd, mem[lc], mem[lc+1]);
                        wordtobytes (wd2, mem[lc+2], mem[lc+3]);
                        lc = lc + 4;
                    }
                    break;
                case skipdir:
                    {
                        scanint(wd);
                        icr.wdf = wd;
                        i1 = icr.pad.sintf;
                        if (i1 < 0) 
                            warn(badskip);
                        else
                            lc = lc + i1;
                    }
                    break;
                case stringdir:
                    {
                        scanstr();
                    }
                    break;
                case labeldir:
                    {
                       getch();
                        if (set::of(range('a','z'), range('A','Q'), range('S','Z'),'$', eos).has(ch))  {
                            scanname(id);
                            idrec = findname(id);
                            if (idrec == nil) 
                                idrec = inname(id);
                            else {
                                if (idrec->loc >= 0) 
                                    warn(namedefined);
                                i1 = idrec->patch;
                                icr.pad.sintf = lc;
                              wordtobytes (icr.wdf, b1, b2);
                                while (i1 >= 0 && i1+1 < MAXINT)  {
//   output<<"label " << i1<<"\n";
                                   bytestoword (mem[i1], mem[i1+1], icr.wdf);
//output << "check\n";
                                    i2 = icr.pad.sintf;
                                    mem[i1] = b1;
                                    mem[i1+1] = b2;
                                    i1 = i2;
                                }   /*while*/
                            }    /*else*/
                            idrec->patch = -1;
                            idrec->loc = lc;
                        } else
            error = invalidname;
//output<<"end label\n";
      }
      break;
     }   /*case of op*/
 }   /*if Error = noerr*/
    }    /*procedure proline*/

    void initopcodes()
    {
 inop = word1::of(eos);
 iaop = word1::of(11, eos);
 isop = word1::of(12, eos);
 imop = word1::of(11, 12, eos);
 idop = word1::of(13, eos);
 fnop = word1::of(11, 13, eos);
 faop = word1::of(12, 13, eos);
 fsop = word1::of(11, 12, 13, eos);
 fmop = word1::of(14, eos);
 fdop = word1::of(11, 14, eos);
 biop = word1::of(12, 14, eos);
 boop = word1::of(11, 12, 14, eos);
 baop = word1::of(13, 14, eos);
 icop = word1::of(11, 13, 14, eos);
 fcop = word1::of(12, 13, 14, eos);
 jsrop = word1::of(11, 12, 13, 14, eos);
 bktop = word1::of(15, eos);
 ldop = word1::of(11, 15, eos);
 stoop = word1::of(12, 15, eos);
 ldaop = word1::of(11, 12, 15, eos);
 fltop = word1::of(13, 15, eos);
 fixop = word1::of(11, 13, 15, eos);
 jop = word1::of(12, 13, 15, eos);
 srop = word1::of(11, 12, 13, 15, eos);
 slop = word1::of(14, 15, eos);
 rdop = word1::of(11, 14, 15, eos);
 wrop = word1::of(12, 14, 15, eos);
 trngop = word1::of(11, 12, 14, 15, eos);
 haltop = word1::of(11, 12, 13, 14, 15, eos);
    }    /*procedure initopcodes*/

int main(int argc, const char* argv[])
{     /*main*/
    pio_initialize(argc, argv);
   // assign ( infile );
    reset ( infile , "codefile");
    symbols = nil;
    initopcodes();
    line = 1;
    lc = 0;
    errs = false;
    warning = false;
    saved = false;
    listing = false;
    if(argc > 1 && argv[1][0]== '-')
    {	for(int ii = 1; ii < strlen(argv[1]);++ii)
        {	if(argv[1][ii] == 'l' || argv[1][ii]  == 'L')
                listing = true;
        }
    }

//output << sizeof(ccrec)<<" "<<sizeof(bcrec)<<" "<<sizeof(icrec)<<'\n';
    
    if (listing && ! eof ( infile )) 
       {
       output << NL;
       output <<  format(lc,4) << ": ";
       }
    while (! eof ( infile )) 
       {
       getch();
       while (! eof ( infile ) && ((ch == '%') || (ch == ' '))) 
          {
          infile  >> NL;
          if (listing && ! eof ( infile )) 
             {     /* Dana -- added 'and not eof' */
             output << NL;
             output <<  format(lc,4) << ": ";
             }
          getch();
          line = line + 1;
          }
       if (! eof ( infile )) 
          {
          error = noerr;
          morewarn = false;
          proline();
          getch();
          if ((ch != '%') && (ch != ' ')) 
             {
             warn(textfollows);
             while (! eoln ( infile )) 
                {
                infile >> ch;
                if (listing) 
                   output << ch;
                }
          }   /*while*/
          infile  >> NL;
          if (listing && (! eof ( infile ))) 
             {     /* Dana -- added 'and not eof' */
             output << NL;
             output <<  format(lc,4) << ": ";
             }
          if (error != noerr) 
             {
             errs = true;
             output << "ERROR -- ";
             switch (error) {
                case invalidname: output << "Invalid Name in Label Directive"; break;
                case unkopname: output << "Unknown Operation or Directive Name"; break;
                case badstring: output << "Improper String Directive -- Missing \""; break;
                case badgenaddr: output << "Improperly Formed General Address"; break;
                case badregaddr: output << "Register Address Out of Range"; break;
                case badint: output << "Improperly Formed Integer Constant"; break;
                case badreal: output << "Improperly Formed Real Constant"; break;
                case badstr: output << "Improperly Formed String Constant"; break;
                case badname: output << "Improperly Formed Name"; break;
                case illregaddr: output << "Direct Register Address not Permitted"; break;
                case illimedaddr: output << "Immediate Address not Permitted"; break;
                case badshftamt: output << "Shift Amount not in Range";
                break;
                }    /*case Error of*/
             output << " detected on line " << format(line, 1) << NL;
             }
          while (morewarn) 
             printwarn();
          line = line + 1;
          }
       }    /*while*/

    checktab(symbols);

  //  assign ( memf );
    rewrite(memf, "OBJ"); /* write an empty file if necessary */
    if (! errs && ! warning) 
       {
       for( i = 0; i <= lc - 1; i ++)
           {
           bcr.bytef = mem[i];
           memf << bcr.pad.hintf;
//           memf <<'.'<< bcr.pad.p1<<'.';
           }    /*for*/
       }

   close ( infile );
   close ( memf );

    return EXIT_SUCCESS;
}
