#include "ptoc.h"

/***********************************************************/
/*                                                         */
/* ICS 4410                                                */
/* Fall '87                                                */
/*                                                         */
/* Lynn R. Akers, Jr.                                      */
/*                                                         */
/* This program has been modified to run on an IBM PC or   */
/* compatable using TURBO Pascal version 3.0.              */
/*                                                         */
/* It has only been tested for MICRO, so some bugs may     */
/* still exist.                                            */
/*                                                         */
/***********************************************************/


/* */
/*$R+*/


/*#include "dos.h"*/

/*version of the MACC1 interpreter for the MACC2 byte-addressable mach */
/* Author: C.T. Wilkes (from the original MACC1 source by A.B. Maccabe)*/
/*                     Last modified: 84/11/23 CTW                     */
const int MAXINT = 32768;

const integer memsiz = MAXINT;	/*15 bit addresses*/
const integer regsiz = 16;
const integer intsiz = MAXINT;

/*addressing modes*/
const integer dreg = 0;   /*direct register*/
const integer dmem = 1;   /*direct memory*/
const integer indxd = 2;   /*indexed*/
const integer immed = 3;   /*immediate*/
const integer ireg = 4;   /*indirect register*/
const integer imem = 5;   /*indirect memory*/
const integer iindxd = 6;   /*indirect indexed*/
const integer pcrel = 7;   /*Pc relative*/

/*opcodes*/
const integer indent = 0;
const integer ia = 1;
const integer is = 2;
const integer im = 3;
const integer id = 4;
const integer fn = 5;
const integer fa = 6;
const integer fs = 7;
const integer fm = 8;
const integer fd = 9;
const integer bi = 10;
const integer bo = 11;
const integer ba = 12;
const integer ic = 13;
const integer fc = 14;
const integer jsr = 15;
const integer bkt = 16;
const integer ld = 17;
const integer sto = 18;
const integer lda = 19;
const integer flt = 20;
const integer fix = 21;
const integer j = 22;
const integer sr = 23;
const integer sl = 24;
const integer rd = 25;
const integer wr = 26;
const integer trng = 27;
const integer halt2 = 31;

typedef unsigned char regaddr;
const int min_regaddr = 0;
const int max_regaddr = 15;
typedef unsigned memaddr;
const int min_memaddr = 0;
const int max_memaddr = MAXINT;

typedef set_template<16> word1;

typedef set_template<8> byte1;


typedef set_template<32> doubleword; // guess

typedef int wordrange;
const int min_wordrange = -MAXINT;
const int max_wordrange = MAXINT -1;

typedef signed char byterange;
const int min_byterange = -128;
const int max_byterange = 127;
/*
struct icrec {
    union {
       word1 wf; 
       wordrange sintf; 
    };
};

struct bcrec {
   union {
      byte1 bytef; 
      byterange hintf; 
   };
};
*/
//struct icrec {
typedef   union icrec{
      word1 wf; // 4 bytes
      struct pad {
//      wordrange p1;
      wordrange sintf; //2 bytes
      }pad;
   } irec;
//};

typedef union bcrec {
   byte1 bytef; //4 bytes
   struct pad {
    byterange p3, p2, p1; //1 byte each
   	byterange hintf; // 1 byte
  } pad;
} bcrec;



struct rcrec {
   union {
      real rf; 
      struct { byte1 b1, b2, b3, b4, b5, b6; } s1;
   };
};
/*
struct ccrec {
 union {
   struct { char c2, c1; } s_true;
   word1 wf;
 };
};

*/
struct ccrec {
  union {
    word1 wf;  
    struct { char c4,c3,c1, c2; } s_true;//was Char c2,c1
//    struct { char c2,c1,c3, c4; } s_false;//was Char c2,c1
  };                        /*Inverted Bytes!!!*/
};

array<min_memaddr,max_memaddr,byte1> mem; /*Memory*/
array<min_regaddr,max_regaddr,word1> regs; /*Registers*/
memaddr pc;   /*Program Counter*/
word1 ir;   /*Instruction Register*/
boolean stopprogram;   /*Machine halt flag*/
boolean lt, eq, gt;
memaddr oldpc;
icrec icr;
rcrec rcr;
ccrec ccr;
bcrec bcr;
array<1,20,char> arg, memfname, dumpfname;
file<byterange> dumpf;
boolean memimage;   /*true iff there is a nonempty memory image*/
boolean trace, dump;
integer i;
integer regindex;        /* used to zero registers */


    void wordtobytes (word1 w,  byte1& hibyte, byte1& lobyte)
    {
        integer i;

 lobyte = byte1::of(eos);
 hibyte = byte1::of(eos);
 for( i = 0; i <= 7; i ++) {
     if (w.has(i)) 
         lobyte = lobyte + byte1::of(i, eos);
     if (w.has(i+8)) 
         hibyte = hibyte + byte1::of(i, eos);
 }   /*for*/
    }    /*procedure wordtobytes*/


    void bytestoword (byte1 hibyte, byte1 lobyte,  word1& w)
    {
        integer i;

 w = word1::of(eos);
 for( i = 0; i <= 7; i ++) {
     if (lobyte.has(i)) 
  w = w + word1::of(i, eos);
     if (hibyte.has(i)) 
  w = w + word1::of(i+8, eos);
 }   /*for*/
    }    /*procedure bytestoword*/

    void wordstodoubleword ( word1 loword, word1 hiword, doubleword& dw)
    {
        integer i;

        dw = doubleword::of(eos);
        for( i = 0; i <= 15; i ++) {
            if (loword.has(i)) 
                dw = dw + doubleword::of(i, eos);
            if (hiword.has(i)) 
                dw = dw + doubleword::of(i+16, eos);
        }   /*for*/
    }    /*procedure wordstodoubleword*/

    void doublewordtowords ( doubleword dw, word1& loword, word1& hiword)
    {
        integer i;

 loword = word1::of(eos);
 hiword = word1::of(eos);
 for( i = 0; i <= 15; i ++) {
     if (dw.has(i)) 
         loword = loword + word1::of(i, eos);
     if (dw.has(i+16)) 
         hiword = hiword + word1::of(i, eos);
 }   /*for*/
    }    /*procedure doublewordtowords*/

    void writeinst (integer loc, integer opcode, integer field1, integer field2, integer field3, integer addr2);

    /*----------------------------------------------*
    * writeinst -                                   *
    *     writes out an instruction to the listing  *
    *-----------------------------------------------*/

static void showreg (integer reg)
   /* write a register symbol to the listing */
{
   output << 'r' << format(reg,1);
}           /* showreg */

    static void showaddr (integer location, integer mode, integer r2, integer a2);

          /* write a symbolic general address to the listing */

static void showindxd (integer r, integer a)
   /* write a symbolic indexed address to the listing */
{
   output << format(a,1) << '(';
   showreg (r);
   output << ')';
}              /* showindxd */



static void showaddr (integer location, integer mode, integer r2, integer a2)

{     /* showaddr */
   switch (mode) {
      case dreg :
         showreg (r2);
         break;
      case dmem :
         output << format(a2,1);
         break;
      case indxd :
         showindxd (r2, a2);
         break;
      case immed :
         output << '#' << format(a2,1);
         break;
      case ireg :
         {
            output << '*';
            showreg (r2);
         }
         break;
      case imem :
         output << '*' << format(a2,1);
         break;
      case iindxd :
         {
            output << '*';
            showindxd (r2, a2);
         }
         break;
      case pcrel :
         output << '&' << format(a2,1);
         break;
      default:
         output << "invalid addressing mode = " << format(mode,1);
   }   /* case */
}           /* showaddr */

    void writeinst (integer loc, integer opcode, integer field1, integer field2, integer field3, integer addr2)

    {     /* writeinst */
       output << format(' ',5) << format(loc,5) << format(' ',5);
       switch (opcode) {
          case indent   : output << "in    "; break;
          case ia   : output << "ia    "; break;
          case is   : output << "is    "; break;
          case im   : output << "im    "; break;
          case id   : output << "id    "; break;

          case fn   : output << "fn    "; break;
          case fa   : output << "fa    "; break;
          case fs   : output << "fs    "; break;
          case fm   : output << "fm    "; break;
          case fd   : output << "fd    "; break;

          case bi   : output << "bi    "; break;
          case bo   : output << "bo    "; break;
          case ba   : output << "ba    "; break;

          case ic   : output << "ic    "; break;
          case fc   : output << "fc    "; break;

          case jsr  : output << "jsr   "; break;

          case bkt  : output << "bkt   "; break;
          case ld   : output << "ld    "; break;
          case sto  : output << "sto   "; break;
          case lda  : output << "lda   "; break;

          case flt  : output << "flt   "; break;
          case fix  : output << "fix   "; break;

          case j    : switch (field1) {
                   case 0 : output << "jmp   "; break;
                   case 1 : output << "jlt   "; break;
                   case 2 : output << "jle   "; break;
                   case 3 : output << "jeq   "; break;
                   case 4 : output << "jne   "; break;
                   case 5 : output << "jge   "; break;
                   case 6 : output << "jgt   "; break;
                   case 7 : output << "nop   "; break;
                default:
                      output << "invalid jump instruction, type = " << 
        format(field1,1) << ": ";
                 }
                 break; /* case */

          case sr   : switch (field2) {
                   case 0 : output << "srz   "; break;
                   case 1 : output << "sro   "; break;
                   case 2 : output << "sre   "; break;
                   case 3 : output << "src   "; break;
                   case 4 : output << "srcz  "; break;
                   case 5 : output << "srco  "; break;
                   case 6 : output << "srce  "; break;
                   case 7 : output << "srcc  "; break;
                   default:
                      output << "invalid sr instruction, mode = " << format(field2,1) << ": ";
                 }
                 break; /* case */

          case sl   : switch (field2) {
                   case 0 : output << "slz   "; break;
                   case 1 : output << "slo   "; break;
                   case 2 : output << "sle   "; break;
                   case 3 : output << "slc   "; break;
                   case 4 : output << "slcz  "; break;
                   case 5 : output << "slco  "; break;
                   case 6 : output << "slce  "; break;
                   case 7 : output << "slcc  "; break;
                   default:
                      output << "invalid sl instruction, mode = " << format(field2,1) << ": ";
                 }
                 break; /* case */

          case rd   : switch (field1) {
                   case 0 : output << "rdi   "; break;
                   case 1 : output << "rdf   "; break;
                   case 2 : output << "rdbd  "; break;
                   case 3 : output << "rdbw  "; break;
                   case 4 : output << "rdod  "; break;
                   case 5 : output << "rdow  "; break;
                   case 6 : output << "rdhd  "; break;
                   case 7 : output << "rdhw  "; break;
                   case 8 : output << "rdch  "; break;
                   case 9 : output << "rdst  "; break;
                  case 10 : output << "rdin  "; break;
                  case 11 : output << "rdln  "; break;
                   default:
                      output << "invalid rd instruction, fmt = " << format(field1,1) << ": ";
                 }
                 break; /* case */

          case wr   : switch (field1) {
                   case 0 : output << "wri   "; break;
                   case 1 : output << "wrf   "; break;
                   case 2 : output << "wrbd  "; break;
                   case 3 : output << "wrbw  "; break;
                   case 4 : output << "wrod  "; break;
                   case 5 : output << "wrow  "; break;
                   case 6 : output << "wrhd  "; break;
                   case 7 : output << "wrhw  "; break;
                   case 8 : output << "wrch  "; break;
                   case 9 : output << "wrst  "; break;
                  case 11 : output << "wrnl  "; break;
                   default:
                      output << "invalid wr instruction, fmt = " << format(field1,1) << ": ";
                 }
                 break; /* case */

          case trng : output << "trng  "; break;
          case halt2 : output << "halt  "; break;

          default:
             output << "invalid instruction, opcode = " << format(opcode,1) << ": ";
       }    /* case */

       if (set::of(range(indent , flt), trng, eos).has(opcode))          /* 2 address */
          {
             showreg (field1);
             output << ", ";
             showaddr (loc, field2, field3, addr2);
          }   /* if */
       else if (((opcode == j) && (field1 != 7))   /* jump but not nop */
            || ((opcode == wr) && (field1 != 11)) /* write but not wrnl */
            || (opcode == rd)) 
          showaddr (loc, field2, field3, addr2);
       else if (set::of(sr, sl, eos).has(opcode))        /* 1 register address */
          {
             showreg (field1);
             output << ", ";
             output << field3;  /* amnt */
          }   /* else if */

       output << NL;
    }    /* writeinst */


    void writeboolean (boolean b)
    {
    if (b) 
       output << "TRUE";
    else
       output << "FALSE";
    }    /* writeboolean */

    void displayregs()
    {
        integer i;

    output << "########################################" << NL;
    output << "  Pc = " << format(pc,1) << "  Lt = ";
    writeboolean (lt);
    output << "  Eq = ";
    writeboolean (eq);
    output << "  Gt = ";
    writeboolean (gt);
    output << NL;
    for( i = 0; i <= regsiz - 1; i ++) {
        icr.wf = regs[i];
        output << "  Reg[" << format(i,2) << "] = " << format(icr.pad.sintf , 5);
        if ((i + 1) % 4 == 0) 
           output << NL;
        }    /*for*/
    output << "########################################" << NL;
    }    /* displayregs */


    void fetchword (word1& w)
    {
    bytestoword (mem[pc], mem[pc+1], w);
    pc = (pc + 2) % memsiz;
    }    /*procedure fetchword*/


    integer eac(integer mode, regaddr reg,  integer& w2)
    /*perform effective address calculation -- update Pc as needed for 2
    word instructions.*/
    {
       integer taddr;     /*temporary address*/

 integer eac_result;
 w2 = 0;
 switch (mode) {
     case dreg: taddr = -1; break;			/*indicates direct reg*/
     case dmem: {
           fetchword (icr.wf);
           /*Icr.wf := Icr.wf - [15];*/
           taddr = icr.pad.sintf;
           w2 = taddr;
           }
           break;
    case indxd:   {
             fetchword (icr.wf);
             w2 = icr.pad.sintf;
             icr.wf = regs[reg];
             taddr = (w2 + icr.pad.sintf) % memsiz;
             }
             break;
    case immed:   {
             taddr = pc;
             bytestoword (mem[taddr], mem[taddr+1], icr.wf);
             w2 = icr.pad.sintf;
             pc = (pc + 2) % memsiz;
             }
             break;
     case ireg: {
           icr.wf = regs[reg]  /*- [15]*/;
           taddr = icr.pad.sintf;
           }
           break;
     case imem: {
           fetchword (icr.wf);
           /*Icr.wf := Icr.wf - [15];*/
           w2 = icr.pad.sintf;
           bytestoword (mem[w2], mem[w2+1], icr.wf);
           /*Icr.wf := Icr.wf - [15];*/
           taddr = icr.pad.sintf;
           }
           break;
   case iindxd: {
           fetchword (icr.wf);
           w2 = icr.pad.sintf;
           icr.wf = regs[reg];
           taddr = (w2 + icr.pad.sintf ) % memsiz;
           bytestoword (mem[taddr], mem[taddr+1], icr.wf);
           /*Icr.wf := Icr.wf - [15];*/
           taddr = icr.pad.sintf;
           }
           break;
    case pcrel: {
           fetchword (icr.wf);
           w2 = icr.pad.sintf;
           taddr = (w2 + pc) % memsiz;
           }
           break;
    }    /*case of mode*/
 eac_result = taddr;
 return eac_result;
    } /*function eac*/


    void twoaddr(integer opcode, integer r1, integer amode, integer r2,  integer& w2)
    /*perform two address operation*/
    {
        integer addr;
        word1 wd;
        integer iop1, iop2, ans;
        real rop1, rop2;
        integer i;

        addr = eac(amode, r2, w2);
        if (set::of(indent, ia, is, im, id, ic, eos).has(opcode))  {
            if (addr >= 0) 
                bytestoword (mem[addr], mem[addr+1], icr.wf);
            else
                icr.wf = regs[r2];
            iop2 = icr.pad.sintf;
            icr.wf = regs[r1];
            iop1 = icr.pad.sintf;
            switch (opcode) {
                case indent:
                    ans = -iop2;
                    break;
                case ia:
                    ans = iop1 + iop2;
                    break;
                case is:
                    ans = iop1 - iop2;
                    break;
                case im:
                    ans = iop1 * iop2;
                    break;
                case id:
                    ans = iop1 / iop2;
                    break;
                case ic:
                    {
                        ans = iop1;     /*to restore state of Regs[r1]*/
                        lt = false;
                        eq = false;
                        gt = false;
                        if (iop1 < iop2) 
                            lt = true;
                        else if (iop1 == iop2) 
                            eq = true;
                        else
                            gt = true;
                    }
                    break;
            }    /*case of opcode*/
            icr.pad.sintf = ans;
            regs[r1] = icr.wf;
        } else if (set::of(fn, fa, fs, fm, fd, fc, eos).has(opcode))  {
           wordtobytes ( regs[r1], rcr.s1.b1, rcr.s1.b4 );
           wordtobytes ( regs[(r1 + 1) % regsiz], rcr.s1.b5, rcr.s1.b6 );
           rcr.s1.b2 = byte1::of(eos);
           rcr.s1.b3 = byte1::of(eos);
            rop1 = rcr.rf;
            if (addr >= 0)  {
        rcr.s1.b1 = mem [ addr     ];
        rcr.s1.b4 = mem [ addr + 1 ];
        rcr.s1.b5 = mem [ addr + 2 ];
        rcr.s1.b6 = mem [ addr + 3 ];
        rcr.s1.b2 = byte1::of(eos);
        rcr.s1.b3 = byte1::of(eos);
            } else {
           wordtobytes ( regs[r2], rcr.s1.b1, rcr.s1.b4 );
           wordtobytes ( regs[(r2 + 1) % regsiz], rcr.s1.b5, rcr.s1.b6 );
           rcr.s1.b2 = byte1::of(eos);
           rcr.s1.b3 = byte1::of(eos);
            }
            rop2 = rcr.rf;
            switch (opcode) {
                case fn:
                    rop1 = -rop2;
                    break;
                case fa:
                    rop1 = rop1 + rop2;
                    break;
                case fs:
                    rop1 = rop1 - rop2;
                    break;
                case fm:
                    rop1 = rop1 * rop2;
                    break;
                case fd:
                    rop1 = rop1 / rop2;
                    break;
                case fc:
                    {
                        lt = false;
                        eq = false;
                        gt = false;
                        if (rop1 < rop2) 
                            lt = true;
                        else if (rop1 == rop2) 
                            eq = true;
                        else
                            gt = true;
                    }
                    break;
            }    /*case of opcode*/
            rcr.rf = rop1;
     bytestoword ( rcr.s1.b1, rcr.s1.b4, regs[r1] );
     bytestoword ( rcr.s1.b5, rcr.s1.b6, regs[(r1 + 1) % regsiz] );
        } else if (set::of(bi, bo, ba, eos).has(opcode))  {
            if (addr >= 0) 
                bytestoword (mem[addr], mem[addr+1], wd);
            else
                wd = regs[r2];
            switch (opcode) {
                case bi:
                    {
                        regs[r1] = word1::of(eos);
                        for( i = 0; i <= 15; i ++)
                            if (! (wd.has(i))) 
                                regs[r1] = regs[r1] + word1::of(i, eos);
                    }
                    break;
                case bo:
                    regs[r1] = regs[r1] + wd;
                    break;
                case ba:
                    regs[r1] = regs[r1] * wd;
                    break;
            }   /*case of opcode*/
        } else if (opcode == jsr)  {
            icr.pad.sintf = pc;
            regs[r1] = icr.wf;
            if (addr >= 0) 
                pc = addr;
            else {
                output << "JSR to a Register " << format(pc - 2, 1) << NL;
                stopprogram = true;
            }
        } else if (opcode == bkt) 
            if (addr < 0)  {
                output << "Address of BKT is a Register " << format(pc - 2, 1) << NL;
                stopprogram = true;
            } else {
                icr.wf = regs[r1];
                iop2 = icr.pad.sintf;
                icr.wf = regs[(r1 + 1) % regsiz];
                for( i = 0; i <= icr.pad.sintf  - 1; i ++)
                    mem[(addr + i) % memsiz] = mem[(iop2 + i) % memsiz];
            }   /*if*/
        else if (opcode == ld) 
            if (addr >= 0) 
                bytestoword (mem[addr], mem[addr+1], regs[r1]);
            else 
                regs[r1] = regs[r2];
        else if (opcode == sto)  
            if (addr >= 0)  {
                wordtobytes (regs[r1], mem[addr], mem[addr+1]);
                if (trace)  {
                    bytestoword (mem[addr], mem[addr+1], icr.wf);
                    output << "***** value " << format(icr.pad.sintf ,1) << 
                             " stored at location " << format(addr,1) << NL;
                }
            } else
                regs[r2] = regs[r1];
        else if (opcode == lda)  
            if (addr >= 0)  {
                icr.pad.sintf = addr;
                regs[r1] = icr.wf;
            } else {
                output << "Address of LDA is a Register " << format(pc - 2, 1) << NL;
                stopprogram = true;
            }
        else if (opcode == flt)  {
            if (addr >= 0) 
                bytestoword (mem[addr], mem[addr+1], icr.wf);
            else
                icr.wf = regs[r2];
            rcr.rf = icr.pad.sintf;
     bytestoword ( rcr.s1.b1, rcr.s1.b4, regs[r1] );
     bytestoword ( rcr.s1.b5, rcr.s1.b6, regs[(r1 + 1) % regsiz] );
        } else if (opcode == fix)  {
            if (addr >= 0)  {
        rcr.s1.b1 = mem [ addr     ];
        rcr.s1.b4 = mem [ addr + 1 ];
        rcr.s1.b5 = mem [ addr + 2 ];
        rcr.s1.b6 = mem [ addr + 3 ];
        rcr.s1.b2 = byte1::of(eos);
        rcr.s1.b3 = byte1::of(eos);
            } else {
           wordtobytes ( regs[r2], rcr.s1.b1, rcr.s1.b4 );
           wordtobytes ( regs[(r2 + 1) % regsiz], rcr.s1.b5, rcr.s1.b6 );
           rcr.s1.b2 = byte1::of(eos);
           rcr.s1.b3 = byte1::of(eos);
            }    /*if*/
            icr.pad.sintf = round(rcr.rf);
            regs[r1] = icr.wf;
        } else if (opcode == trng) 
            if (addr >= 0)  {
                bytestoword (mem[addr], mem[addr+1], icr.wf);
                iop1 = icr.pad.sintf;		/*lower bound*/
                bytestoword (mem[addr+2], mem[addr+3], icr.wf);
                iop2 = icr.pad.sintf;		/*upper bound*/
                icr.wf = regs[r1];
                ans = icr.pad.sintf;		/*test expression*/
/* ***** On next line, Icr.pad.sintf should be ans?! ***** */
                if ((ans < iop1) || (ans > iop2))  {
                   output << "Index expression out of bounds at " << format(pc-2,1) << 
                           " (expr = " << format(ans,1) << " lower = " << format(iop1,1) << 
                           " upper = " << format(iop2,1) << ')' << NL;
                   stopprogram = true;
                }
            } else {
                output << "Address of TRNG is a register at " << format(pc-2,1) << NL;
                stopprogram = true;
            }
    }    /*procedure twoaddr*/


    void jop(integer jmode, integer amode, integer reg,  integer& w2)
    /*perform jump operation*/
    {
        integer addr;

        addr = eac(amode, reg, w2);
        if (addr < 0)  {
            output << "Jump to a Register " << format(pc - 2, 1) << NL;
            stopprogram = true;
        } else if (set::of(0, 1, 2, 3, 4, 5, 6, 7, eos).has(jmode))  
            switch (jmode) {
                case 0:
                    pc = addr;
                    break;
                case 1:
                    if (lt) 
                        pc = addr;
                        break;
                case 2:
                    if (lt || eq) 
                        pc = addr;
                        break;
                case 3:
                    if (eq) 
                        pc = addr;
                        break;
                case 4:
                    if (! eq)  
                        pc = addr;
                        break;
                case 5:
                    if (gt || eq)  
                        pc = addr;
                        break;
                case 6:
                    if (gt) 
                        pc = addr;
                        break;
                case 7:/*null*/;
                break;
            }   /*case of jmode*/
        else {
            output << "Invalid Jump Mode " << format(pc - 2, 1) << NL;
            stopprogram = true;
        }   /*if*/
    }    /*procedure jop*/


    void srop(integer reg, integer smode, integer amnt)
    /*perform right shift*/
    {
        word1 w1, w2, w3;
        integer i, j;
        boolean flag;

        switch (smode) {
            case 0:
                {
                    w1 = regs[reg];
                    if (amnt == 0) 
                        w1 = word1::of(eos);
                    else
                        for( i = 1; i <= amnt; i ++) {
                            for( j = 0; j <= 15 - i; j ++) {
                                w1 = w1 - word1::of(j, eos);
                                if (w1.has(j + 1)) 
                                    w1 = w1 + word1::of(j, eos);
                            }    /*for j*/
                            w1 = w1 - word1::of(16 - i, eos);
                        }    /*for i*/
                    regs[reg] = w1;
                }
                break;
            case 1:
                {
                    w1 = regs[reg];
                    if (amnt == 0) 
                        w1 = word1::of(range(0,15), eos);
                    else
                        for( i = 1; i <= amnt; i ++) {
                            for( j = 0; j <= 15 - i; j ++) {
                                w1 = w1 - word1::of(j, eos);
                                if (w1.has(j + 1)) 
                                    w1 = w1 + word1::of(j, eos);
                            }    /*for j*/
                            w1 = w1 + word1::of(16 - i, eos);
                        }    /*for i*/
                    regs[reg] = w1;
                }
                break;
            case 2:
                {
                    w1 = regs[reg];
                    if (amnt == 0)  
                        if (w1.has(15)) 
                            w1 = word1::of(range(0,15), eos);
                        else 
                            w1 = word1::of(eos);
                    else
                        for( i = 1; i <= amnt; i ++)
                            for( j = 0; j <= 14; j ++) {
                                w1 = w1 - word1::of(j, eos);
                                if (w1.has(j + 1)) 
                                    w1 = w1 + word1::of(j, eos);
                            }    /*for j*/
                    regs[reg] = w1;
                }
                break;
            case 3:
                {				/*SRC*/
                    w1 = regs[reg];
                    if (amnt != 0)  
                        for( i = 1; i <= amnt; i ++) {
                            flag = w1.has(0);
                            for( j = 0; j <= 14; j ++) {
                                w1 = w1 - word1::of(j, eos);
                                if (w1.has(j + 1)) 
                                    w1 = w1 + word1::of(j, eos);
                            }    /*for j*/
                            if (flag) 
                                w1 = w1 + word1::of(15, eos);
                            else 
                                w1 = w1 - word1::of(15, eos);
                        }    /*for i*/
                    regs[reg] = w1;
                }
                break;
            case 4:
                {				/*SRCZ*/
                    w1 = regs[reg];
                    w2 = regs[(reg + 1) % regsiz];
                    if (amnt == 0)  {
                        w2 = w1;
                        w1 = word1::of(eos);
                    } else {
                        for( i = 1; i <= amnt; i ++) {
                            for( j = 0; j <= 14; j ++) {
                                w2 = w2 - word1::of(j, eos);
                                if (w2.has(j + 1)) 
                                    w2 = w2 + word1::of(j, eos);
                            }    /*for j*/
                            if (w1.has(0))  
                                w2 = w2 + word1::of(15, eos);
                            else
                                w2 = w2 - word1::of(15, eos);
                            for( j = 0; j <= 15 - i; j ++) {
                                w1 = w1 - word1::of(j, eos);
                                if (w1.has(j + 1)) 
                                    w1 = w1 + word1::of(j, eos);
                            }    /*for j*/
                            w1 = w1 - word1::of(16 - i, eos);
                        }
                    }   /*for i*/ /*else*/
                    regs[reg] = w1;
                    regs[(reg + 1) % regsiz] = w2;
                }
                break;
            case 5:
                {				/*SRCO*/
                    w1 = regs[reg];
                    w2 = regs[(reg + 1) % regsiz];
                    if (amnt == 0)  {
                        w2 = w1;
                        w1 = word1::of(range(0,15), eos);
                    } else {
                        for( i = 1; i <= amnt; i ++) {
                            for( j = 0; j <= 14; j ++) {
                                w2 = w2 - word1::of(j, eos);
                                if (w2.has(j + 1))  
                                    w2 = w2 + word1::of(j, eos);
                            }    /*for j*/
                            if (w1.has(0)) 
                                w2 = w2 + word1::of(15, eos);
                            else 
                                w2 = w2 - word1::of(15, eos);
                            for( j = 0; j <= 15 - i; j ++) {
                                w1 = w1 - word1::of(j, eos);
                                if (w1.has(j + 1)) 
                                    w1 = w1 + word1::of(j, eos);
                            }    /*for j*/
                            w1 = w1 + word1::of(16 - i, eos);
                        }
                    }   /*for i*/ /*else*/
                    regs[reg] = w1;
                    regs[(reg + 1) % regsiz] = w2;
                }
                break;
            case 6:
                {
                    w1 = regs[reg];
                    w2 = regs[(reg + 1) % regsiz];
                    if (amnt == 0)  {
                        w2 = w1;
                        if (w2.has(15))  
                            w1 = word1::of(range(0,15), eos);
                        else
                            w1 = word1::of(eos);
                    } else {
                        for( i = 1; i <= amnt; i ++) {
                            for( j = 0; j <= 14; j ++) {
                                w2 = w2 - word1::of(j, eos);
                                if (w2.has(j + 1)) 
                                    w2 = w2 + word1::of(j, eos);
                            }    /*for j*/
                            if (w1.has(0))  
                                w2 = w2 + word1::of(15, eos);
                            else 
                                w2 = w2 - word1::of(15, eos);
                            for( j = 0; j <= 15 - i; j ++) {
                                w1 = w1 - word1::of(j, eos);
                                if (w1.has(j + 1)) 
                                    w1 = w1 + word1::of(j, eos);
                            }
                        }   /*for j*/
                    }   /*for i*/ /*else*/
                    regs[reg] = w1;
                    regs[(reg + 1) % regsiz] = w2;
                }
                break;
            case 7:
                {
                    w1 = regs[reg];
                    w2 = regs[(reg + 1) % regsiz];
                    if (amnt == 0)  {
                        w3 = w2;
                        w2 = w1;
                        w1 = w3;
                    } else {
                        for( i = 1; i <= amnt; i ++) {
                            flag = w2.has(0);
                            for( j = 0; j <= 14; j ++) {
                                w2 = w2 - word1::of(j, eos);
                                if (w2.has(j + 1))  
                                    w2 = w2 + word1::of(j, eos);
                            }    /*for j*/
                            if (w1.has(0)) 
                                w2 = w2 + word1::of(15, eos);
                            else
                                w2 = w2 - word1::of(15, eos);
                            for( j = 0; j <= 14; j ++) {
                                w1 = w1 - word1::of(j, eos);
                                if (w1.has(j + 1)) 
                                    w1 = w1 + word1::of(j, eos);
                            }    /*for j*/
                            if (flag)  
                                w1 = w1 + word1::of(15, eos);
                            else
                                w1 = w1 - word1::of(15, eos);
                        }
                    }   /*for i*/ /*else*/
                    regs[reg] = w1;
                    regs[(reg + 1) % regsiz] = w2;
                }
                break;
        }   /*case of smode*/
    }    /*procedure srop*/


    void slop(integer reg, integer smode, integer amnt)
    /*perform left shift*/
    {
        word1 w1, w2, w3;
        integer i, j;
        boolean flag;

        switch (smode) {
            case 0:
                {
                    w1 = regs[reg];
                    if (amnt == 0) 
                        w1 = word1::of(eos);
                    else 
                        for( i = 1; i <= amnt; i ++) {
                            for( j = 14; j >= i - 1; j --) {
                                w1 = w1 - word1::of(j + 1, eos);
                                if (w1.has(j))  
                                    w1 = w1 + word1::of(j + 1, eos);
                            }    /*for j*/
                            w1 = w1 - word1::of(i - 1, eos);
                        }    /*for i*/
                    regs[reg] = w1;
                }
                break;
            case 1:
                {
                    w1 = regs[reg];
                    if (amnt == 0) 
                        w1 = word1::of(range(0,15), eos);
                    else 
                        for( i = 1; i <= amnt; i ++) {
                            for( j = 14; j >= i - 1; j --) {
                                w1 = w1 - word1::of(j + 1, eos);
                                if (w1.has(j))  
                                    w1 = w1 + word1::of(j + 1, eos);
                            }    /*for j*/
                            w1 = w1 + word1::of(i - 1, eos);
                        }    /*for i*/
                    regs[reg] = w1;
                }
                break;
            case 2:
                {
                    w1 = regs[reg];
                    if (amnt == 0) 
                        if (w1.has(0))  
                            w1 = word1::of(range(0,15), eos);
                        else
                            w1 = word1::of(eos);
                    else 
                        for( i = 1; i <= amnt; i ++)
                            for( j = 14; j >= i - 1; j --) {
                                w1 = w1 - word1::of(j + 1, eos);
                                if (w1.has(j))  
                                    w1 = w1 + word1::of(j + 1, eos);
                            }    /*for j*/
                    regs[reg] = w1;
                }
                break;
            case 3:
                {
                    w1 = regs[reg];
                    if (amnt != 0) 
                        for( i = 1; i <= amnt; i ++) {
                            flag = w1.has(15);
                            for( j = 14; j >= 0; j --) {
                                w1 = w1 - word1::of(j + 1, eos);
                                if (w1.has(j))  
                                    w1 = w1 + word1::of(j + 1, eos);
                            }    /*for j*/
                            if (flag) 
                                w1 = w1 + word1::of(0, eos);
                            else
                                w1 = w1 - word1::of(0, eos);
                        }    /*for i*/
                    regs[reg] = w1;
                }
                break;
            case 4:
                {
                    w1 = regs[reg];
                    w2 = regs[(reg + 1) % regsiz];
                    if (amnt == 0)  {
                        w1 = w2;
                        w2 = word1::of(eos);
                    } else {
                        for( i = 1; i <= amnt; i ++) {
                            for( j = 14; j >= 0; j --) {
                                w1 = w1 - word1::of(j + 1, eos);
                                if (w1.has(j))  
                                    w1 = w1 + word1::of(j + 1, eos);
                            }    /*for j*/
                            if (w2.has(15)) 
                                w1 = w1 + word1::of(0, eos);
                            else
                                w1 = w1 - word1::of(0, eos);
                            for( j = 14; j >= i - 1; j --) {
                                w2 = w2 - word1::of(j + 1, eos);
                                if (w2.has(j)) 
                                    w2 = w2 + word1::of(j + 1, eos);
                            }    /*for j*/
                            w2 = w2 - word1::of(i - 1, eos);
                        }
                    }   /*for i*/ /*else*/
                    regs[reg] = w1;
                    regs[(reg + 1) % regsiz] = w2;
                }
                break;
            case 5:
                {
                    w1 = regs[reg];
                    w2 = regs[(reg + 1) % regsiz];
                    if (amnt == 0)  {
                        w1 = w2;
                        w2 = word1::of(range(0,15), eos);
                    } else {
                        for( i = 1; i <= amnt; i ++) {
                            for( j = 14; j >= 0; j --) {
                                w1 = w1 - word1::of(j + 1, eos);
                                if (w1.has(j)) 
                                    w1 = w1 + word1::of(j + 1, eos);
                            }    /*for j*/
                            if (w2.has(15)) 
                                w1 = w1 + word1::of(0, eos);
                            else 
                                w1 = w1 - word1::of(0, eos);
                            for( j = 14; j >= i - 1; j --) {
                                w2 = w2 - word1::of(j + 1, eos);
                                if (w2.has(j)) 
                                    w2 = w2 + word1::of(j + 1, eos);
                            }    /*for j*/
                            w2 = w2 + word1::of(i - 1, eos);
                        }
                    }   /*for i*/ /*else*/
                    regs[reg] = w1;
                    regs[(reg + 1) % regsiz] = w2;
                }
                break;
            case 6:
                {
                    w1 = regs[reg];
                    w2 = regs[(reg + 1) % regsiz];
                    if (amnt == 0)  {
                        w1 = w2;
                        if (w2.has(0)) 
                            w2 = word1::of(range(0,15), eos);
                        else
                            w2 = word1::of(eos);
                    } else {
                        for( i = 1; i <= amnt; i ++) {
                            for( j = 14; j >= 0; j --) {
                                w1 = w1 - word1::of(j + 1, eos);
                                if (w1.has(j)) 
                                    w1 = w1 + word1::of(j + 1, eos);
                            }    /*for j*/
                            if (w2.has(15)) 
                                w1 = w1 + word1::of(0, eos);
                            else
                                w1 = w1 - word1::of(0, eos);
                            for( j = 14; j >= i - 1; j --) {
                                w2 = w2 - word1::of(j + 1, eos);
                                if (w2.has(j)) 
                                    w2 = w2 + word1::of(j + 1, eos);
                            }
                        }   /*for j*/
                    }   /*for i*/ /*else*/
                    regs[reg] = w1;
                    regs[(reg + 1) % regsiz] = w2;
                }
                break;
            case 7:
                {
                    w1 = regs[reg];
                    w2 = regs[(reg + 1) % regsiz];
                    if (amnt == 0)  {
                        w3 = w1;
                        w1 = w2;
                        w2 = w3;
                    } else {
                        for( i = 1; i <= amnt; i ++) {
                            flag = w1.has(15);
                            for( j = 14; j >= 0; j --) {
                                w1 = w1 - word1::of(j + 1, eos);
                                if (w1.has(j)) 
                                    w1 = w1 + word1::of(j + 1, eos);
                            }    /*for j*/
                            if (w2.has(15))  
                                w1 = w1 + word1::of(0, eos);
                            else 
                                w1 = w1 - word1::of(0, eos);
                            for( j = 14; j >= 0; j --) {
                                w2 = w2 - word1::of(j + 1, eos);
                                if (w2.has(j))  
                                    w2 = w2 + word1::of(j + 1, eos);
                            }    /*for j*/
                            if (flag)  
                                w2 = w2 + word1::of(0, eos);
                            else
                                w2 = w2 - word1::of(0, eos);
                        }
                    }   /*for i*/ /*else*/
                    regs[reg] = w1;
                    regs[(reg + 1) % regsiz] = w2;
                }
                break;
        }   /*case of smode*/
    }    /*procedure slop*/


    void rdop(integer rmode, integer amode, integer reg,  integer& w2)
    /*perform a read operation*/
    {
        word1 wd;
        boolean flag;
        integer addr;
        integer i;
        integer ival;
        char ch;

        if (set::of(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, eos).has(rmode))  {
            addr = eac(amode, reg, w2);
            switch (rmode) {
                case 0:
                    {			        /*RDI--read integer*/
                        input >> i;
                        if ((i > MAXINT) || (i < -MAXINT))  {
                            output << "Invalid Input: RDI " << format(pc - 2, 1) << NL;
                            stopprogram = true;
                        } else {
                            icr.pad.sintf = i;
                            if (addr >= 0) 
                                wordtobytes (icr.wf, mem[addr], mem[addr+1]);
                            else
                                regs[reg] = icr.wf;
                        }   /*else*/
                    }
                    break;
                case 1:
                    {			        /*RDF--read floating point*/
                        input >> rcr.rf;
                        if (addr >= 0)  {
      mem [ addr     ] = rcr.s1.b1;
      mem [ addr + 1 ] = rcr.s1.b4;
      mem [ addr + 2 ] = rcr.s1.b5;
      mem [ addr + 3 ] = rcr.s1.b6;
                        } else {
     bytestoword ( rcr.s1.b1, rcr.s1.b4, regs[reg] );
     bytestoword ( rcr.s1.b5, rcr.s1.b6, regs[(reg + 1) % regsiz] );
                        }   /*else*/
                    }
                    break;
                case 2:
                    {			        /*RDBD--read binary digit*/
                        input >> ch;
                        if (ch == '0') 
                            if (addr >= 0) 
                                mem[addr] = mem[addr] - byte1::of(7, eos);
                            else
                                regs[addr] = regs[addr] - word1::of(15, eos);
                        else if (ch == '1') 
                            if (addr >= 0) 
                                mem[addr] = mem[addr] + byte1::of(7, eos);
                            else
                                regs[addr] = regs[addr] + word1::of(15, eos);
                        else {
                            output << "Invalid Input: RDBD " << format(pc - 2, 1) << NL;
                            stopprogram = true;
                        }
                    }
                    break;
                case 3:
                    {			        /*RDBW--read binary word*/
                        i = 15;
                        flag = true;
                        while ((i >= 0) && flag)  {
                            input >> ch;
                            if (ch == '1') 
                                wd = wd + word1::of(i, eos);
                            else if (ch == '0') 
                                wd = wd - word1::of(i, eos);
                            else {
                                output << "Invalid Input: RDBW " << format(pc - 2, 1) << NL;
                                flag = false;
                                stopprogram = true;
                            }    /*else*/
                            i = i - 1;
                        }    /*while*/
                        if (flag)  
                            if (addr >= 0) 
                                wordtobytes (wd, mem[addr], mem[addr+1]);
                            else
                                regs[reg] = wd;
                    }
                    break;
                case 4:                              /*RDOD--read octal digit*/
                case 5:                              /*RDOW--read octal word*/
                    {
                        output << "Operations RDOD and RDOW not presently" << 
                                 "implemented, at " << format(pc-2,1) << NL;
                        stopprogram = true;
                    }
                    break;
                case 6:
                    {			        /*RDHD--read hex digit*/
                        input >> ch;
                        if (set::of(range('A','F'), range('a','f'), range('0','9'), eos).has(ch))  {
                            switch (ch) {
                                case 'A': case 'a':
                                    wd = word1::of(15, 13, eos);
                                    break;
                                case 'B': case 'b':
                                    wd = word1::of(15, 13, 12, eos);
                                    break;
                                case 'C': case 'c':
                                    wd = word1::of(15, 14, eos);
                                    break;
                                case 'D': case 'd':
                                    wd = word1::of(15, 14, 12, eos);
                                    break;
                                case 'E': case 'e':
                                    wd = word1::of(15, 14, 13, eos);
                                    break;
                                case 'F': case 'f':
                                    wd = word1::of(15, 14, 13, 12, eos);
                                    break;
                                case '0':
                                    wd = word1::of(eos);
                                    break;
                                case '1':
                                    wd = word1::of(12, eos);
                                    break;
                                case '2':
                                    wd = word1::of(13, eos);
                                    break;
                                case '3':
                                    wd = word1::of(13, 12, eos);
                                    break;
                                case '4':
                                    wd = word1::of(14, eos);
                                    break;
                                case '5':
                                    wd = word1::of(14, 12, eos);
                                    break;
                                case '6':
                                    wd = word1::of(14, 13, eos);
                                    break;
                                case '7':
                                    wd = word1::of(14, 13, 12, eos);
                                    break;
                                case '8':
                                    wd = word1::of(15, eos);
                                    break;
                                case '9':
                                    wd = word1::of(15, 12, eos);
                                    break;
                            }    /*case of ch*/
                            if (addr >= 0)  {;
                                /*Mem[addr] := Mem[addr] - [15, 14, 13, 12];*/
                                /*Mem[addr] := Mem[addr] + wd*/
                            } else {
                                regs[reg] = regs[reg] - word1::of(15, 14, 13, 12, eos);
                                regs[reg] = regs[reg] + wd;
                            }   /*else*/
                        } else {
                            output << "Invalid Input: RDHD " << format(pc - 2, 1) << NL;
                            stopprogram = true;
                        }   /*else*/
                    }
                    break;
                case 7:
                    {			        /*RDHW--read hex word*/
                        i = 15;
                        flag = true;
                        wd = word1::of(eos);
                        while ((i >= 1) && flag)  {
                            input >> ch;
                            if (word1::of(range('A','F'), eos).has(ch)) 
                                ival = ord(ch) - ord('A') + 10;
                            else if (word1::of(range('a','f'), eos).has(ch)) 
                                ival = ord(ch) - ord('a') + 10;
                            else if (word1::of(range('0','9'), eos).has(ch))  
                                ival = ord(ch) - ord('0');
                            else {
                                output << "Invalid Input: RDHW " << format(pc - 2, 1) << NL;
                                flag = false;
                                stopprogram = true;
                            }   /*else*/
                            if (flag)  {
                                if (ival > 7)  {
                                    wd = wd + word1::of(i, eos);
                                    ival = ival - 7;
                                }
                                i = i - 1;
                                if (ival > 3)  {
                                    wd = wd + word1::of(i, eos);
                                    ival = ival - 3;
                                }
                                i = i - 1;
                                if (ival > 1)  {
                                    wd = wd + word1::of(i, eos);
                                    ival = ival - 1;
                                }
                                i = i - 1;
                                if (ival > 0)  
                                    wd = wd + word1::of(i, eos);
                                i = i - 1;
                            }   /*if*/
                        }    /*while*/
                        if (addr >= 0)  
                            wordtobytes (wd, mem[addr], mem[addr+1]);
                        else
                            regs[reg] = wd;
                    }
                    break;
                case 8:
                    {			        /*RDCH--read ASCII character*/
                        if (addr >= 0)  
                            bytestoword (mem[addr], mem[addr+1], ccr.wf);
                        else 
                            ccr.wf = regs[reg];
                        input >> ccr.s_true.c1;
                        if (addr >= 0) 
                            wordtobytes (ccr.wf, mem[addr], mem[addr+1]);
                        else
                            regs[reg] = ccr.wf;
                    }
                    break;
                case 9:
                    {			        /*RDST--read ASCII String*/
                        if (addr < 0)  {
                            output << "Invalid Address: RDST " << format(pc - 2, 1) << NL;
                            stopprogram = true;
                        } else {
                            flag = true;
                            while (! eoln()) 
                                if (flag)  {
                                    flag = false;
                                    input >> ccr.s_true.c1;
                                } else {
                                    flag = true;
                                    input >> ccr.s_true.c2;
                                    wordtobytes
                                        (ccr.wf, mem[addr], mem[addr+1]);
                                    addr = (addr + 2) % memsiz;
                                }    /*else*/
                            if (flag) 
                                ccr.s_true.c1 = chr(0);
                            else
                                ccr.s_true.c2 = chr(0);
                            wordtobytes (ccr.wf, mem[addr], mem[addr+1]);
                        }    /*else*/
                        input >> NL;
                    }
                    break;
               case 10:                           /*RDIN--read instruction*/
                    {
                        output << "Operation RDIN not presently implemented" << 
                                ", at " << format(pc-2,1) << NL;
                        stopprogram = true;
                    }
                    break;
               case 11:                           /*RDNL--read newline*/
                    input >> NL;
                    break;
            }   /*case of rmode*/
        } else {
            output << "Invalid Read Mode Detected at " << format(pc - 2, 1) << NL;
            stopprogram = true;
        }
    }    /* rdop */


    void wrop(integer wmode, integer amode, integer reg,  integer& w2)
    /*perform a write operation*/
    {
        integer addr;
        integer ival, bval, i;
        word1 wd;
        char ch;

        if (set::of(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11, eos).has(wmode))  {
            addr = eac(amode, reg, w2);
            switch (wmode) {
                case 0:
                    {			        /*WRI*/
                        if (addr >= 0) 
                            bytestoword (mem[addr], mem[addr+1], icr.wf);
                        else
                            icr.wf = regs[reg];
                        output << format(icr.pad.sintf , 1);
                    }
                    break;
                case 1:
                    {			        /*WRF*/
                        if (addr >= 0)  {
        rcr.s1.b1 = mem [ addr     ];
        rcr.s1.b4 = mem [ addr + 1 ];
        rcr.s1.b5 = mem [ addr + 2 ];
        rcr.s1.b6 = mem [ addr + 3 ];
        rcr.s1.b2 = byte1::of(eos);
        rcr.s1.b3 = byte1::of(eos);
                        } else {
           wordtobytes ( regs[reg], rcr.s1.b1, rcr.s1.b4 );
           wordtobytes ( regs[(reg + 1) % regsiz], rcr.s1.b5, rcr.s1.b6 );
           rcr.s1.b2 = byte1::of(eos);
           rcr.s1.b3 = byte1::of(eos);
                        }    /*if*/
                        output << rcr.rf;
                    }
                    break;
                case 2:
                    {			        /*WRBD*/
                        if (addr >= 0) 
                            bytestoword (mem[addr], mem[addr+1], wd);
                        else
                            wd = regs[reg];
                        if (wd.has(15)) 
                            output << '1';
                        else
                            output << '0';
                    }
                    break;
                case 3:
                    {			        /*WRBW*/
                        if (addr >= 0)  
                            bytestoword (mem[addr], mem[addr+1], wd);
                        else 
                            wd = regs[reg];
                        for( i = 0; i <= 15; i ++) 
                            if (wd.has(i)) 
                                output << '1';
                            else 
                                output << '0';
                    }
                    break;
                case 4:                              /*WROD--write octal digit*/
                case 5:                              /*WROW--write octal word*/
                    {
                        output << "Operations WROD and WROW not presently" << 
                                "implemented, at " << format(pc-2,1) << NL;
                        stopprogram = true;
                    }
                    break;
                case 6:
                    {			        /*WRHD*/
                        if (addr >= 0)  
                            bytestoword (mem[addr], mem[addr+1], wd);
                        else 
                            wd = regs[reg];
                        bval = 8;
                        ival = 0;
                        for( i = 15; i >= 12; i --) {
                            if (wd.has(i)) 
                                ival = ival + bval;
                            bval = bval / 2;
                        }    /*for*/
                        output << format(ival, 1) /*hex*/;
                    }
                    break;
                case 7:
                    {			        /*WRHW*/
                        if (addr >= 0) 
                            bytestoword (mem[addr], mem[addr+1], wd);
                        else
                            wd = regs[reg];
                        bval = MAXINT;
                        ival = 0;
                        for( i = 15; i >= 0; i --) {
                            if (wd.has(i)) 
                                ival = ival + bval;
                            bval = bval / 2;
                        }    /*for*/
                        if (ival < 4096) 
                            output << '0';
                        if (ival < 256) 
                            output << '0';
                        if (ival < 16) 
                            output << '0';
                        output << format(ival, 1) /*hex*/;
                    }
                    break;
                case 8:				/*WRCH--write character*/
                    {
                        if (addr >= 0) 
                            bytestoword (mem[addr], mem[addr+1], ccr.wf);
                        else
                            ccr.wf = regs[reg];
                        output << ccr.s_true.c1;
                    }
                    break;
                case 9:				/*WRST--write string*/
                    {
                        ch = ' ';
                        while (ord(ch) != 0)  {
                            if (addr >= 0)  {
                                bytestoword (mem[addr], mem[addr+1], ccr.wf);
                                addr = (addr + 2) % memsiz;
                            } else {
                                ccr.wf = regs[reg];
                                reg = (reg + 1) % regsiz;
                            }    /*if*/
                            if (ord(ccr.s_true.c1) != 0)  {
                                output << ccr.s_true.c1;
                                if (ord(ccr.s_true.c2) != 0) 
                                    output << ccr.s_true.c2;
                                else 
                                    ch = chr(0);
                            } else
                                ch = chr(0);
                        }   /*while*/
                    }
                    break;
               case 11:				/*WRNL--write newline*/
                    output << NL;
                    break;
            }   /*case of wmode*/
        } else {
            output << "Invalid Write Mode Dectected at " << format(pc - 1, 1) << NL;
            stopprogram = true;
        }
    }    /* wrop */


    void execute()
    {
        integer opcode;
        integer f1;
        integer f2;
        integer f3;
        integer w2;

        opcode = 0;
        if (ir.has(15)) 
            opcode = 16;
        if (ir.has(14)) 
            opcode = opcode + 8;
        if (ir.has(13)) 
            opcode = opcode + 4;
        if (ir.has(12))  
            opcode = opcode + 2;
        if (ir.has(11)) 
            opcode = opcode + 1;
        f1 = 0;
        if (ir.has(10))  
            f1 = 8;
        if (ir.has(9)) 
            f1 = f1 + 4;
        if (ir.has(8))  
            f1 = f1 + 2;
        if (ir.has(7)) 
            f1 = f1 + 1;
        f2 = 0;
        if (ir.has(6)) 
            f2 = 4;
        if (ir.has(5))  
            f2 = f2 + 2;
        if (ir.has(4)) 
            f2 = f2 + 1;
        f3 = 0;
        if (ir.has(3)) 
            f3 = 8;
        if (ir.has(2))  
            f3 = f3 + 4;
        if (ir.has(1)) 
            f3 = f3 + 2;
        if (ir.has(0))  
            f3 = f3 + 1;
        w2 = 0;
        switch (opcode) {
            case indent: case ia: case is: case im: case id: case fn: case fa:
            case fs: case fm: case fd: case bi: case bo: case ba: case ic:
            case fc: case jsr: case bkt: case ld: case sto: case lda: case fix:
            case flt: case trng:
                twoaddr(opcode, f1, f2, f3, w2);
                break;
            case j:
                jop(f1, f2, f3, w2);
                break;
            case sr:
                srop(f1, f2, f3);
                break;
            case sl:
                slop(f1, f2, f3);
                break;
            case rd:
                rdop(f1, f2, f3, w2);
                break;
            case wr:
                wrop(f1, f2, f3, w2);
                break;
            case halt2:
                stopprogram = true;
                break;
            case 28: case 29: case 30:
                {
                    output << "Bad opcode" << format(pc - 2, 1) << NL;
                    stopprogram = true;
                }
                break;
        }   /*case*/
        if (trace)  {
           output << "----------------------------------------" << NL;
           writeinst (oldpc, opcode, f1, f2, f3, w2);
           output << "----------------------------------------" << NL;
           displayregs();
        }   /*for*/
    }    /*procedure execute*/

    void rdmem()
    {
       file<byterange> memf;

 integer i;

   // assign ( memf );
    reset(memf, "OBJ");
    if (eof(memf))  memimage = false;
    i = 0;
    while (! eof(memf))  {
        memf >> bcr.pad.hintf;
        mem[i] = bcr.bytef;
        i = i + 1;
    }   /*while*/
    } /*procedure rdmem*/
/*
void getcommand()
{
  integer i,j;
  string s;

  for( i = 1; i <= paramcount; i ++)
  {
    s = paramstr(i);
    if (s[1] == '/') 
      for( j = 2; j <= length(s); j ++)
        switch (upcase(s[j])) {
          case 'D' : dump= true; break;
          case 'T': trace= true; break;
        default: {
          output << "Invalid option: " << s[j] << NL;
        
          }
        }
    else;
        
  }
}
*/


int main(int argc, const char* argv[])
{     /*main program*/
    pio_initialize(argc, argv);
    dump = false;
    trace = false;
    
    if(argc > 1 && argv[1][0]== '-')
    {	for(int ii = 1; ii < strlen(argv[1]);++ii)
        {	if(argv[1][ii] == 'd' || argv[1][ii]  == 'D')
                dump = true;
            else if(argv[1][ii] == 't' || argv[1][ii]  == 'T')
                trace = true;
        }
    }
    memimage = true;
   // getcommand();
    rdmem();
    lt = false;
    eq = false;
    gt = false;
    pc = 0;
    stopprogram = false;
    if (memimage) 
       {
       for( regindex = 0; regindex <= 15; regindex ++)
          regs [ regindex ] = word1::of(eos);
       do {
       oldpc = pc;
       fetchword (ir);         /*fetch the next instruction*/
       execute();
       if (trace) 
          {
          output  << NL;
          output << "Press <ENTER> to contine."  << NL;
          input  >> NL;
          output  << NL;
          }
       } while (!stopprogram);    /*execute the current inst*/
       }
    if (dump) 
       {
      // assign ( dumpf );
       rewrite(dumpf , "DUMPFILE.txt");
       for( i = 1; i <= memsiz - 1; i ++) {
           bcr.bytef = mem[i];
           dumpf << bcr.pad.hintf;
           }   /*for*/
       }   /*if*/
    return EXIT_SUCCESS;
}
