{***********************************************************}
{*                                                         *}
{* ICS 4410                                                *}
{* Fall '87                                                *}
{*                                                         *}
{* Lynn R. Akers, Jr.                                      *}
{*                                                         *}
{* This program has been modified to run on an IBM PC or   *}
{* compatable using TURBO Pascal version 3.0.              *}
{*                                                         *}
{* It has only been tested for MICRO, so some bugs may     *}
{* still exist.                                            *}
{*                                                         *}
{*   Joseph Bergin                                         *}
{*   Modified for the Macintosh, TML Pascal   1989         *}
{*                                                         *}
{*                                                         *}
{*                                                         *}
{***********************************************************}

PROGRAM macc2 (inFile, outFile);
{version of the MACC1 interpreter for the MACC2 byte-addressable mach }
{ Author: C.T. Wilkes (from the original MACC1 source by A.B. Maccabe)}
{                     Last modified: 84/11/23 CTW                     }

	CONST
		MEMSIZ = 16383; {Maxint}
	{15 bit addresses}
		REGSIZ = 16;
		INTSIZ = maxint;

    {addressing modes}
		dreg = 0;   {direct register}
		dmem = 1;   {direct memory}
		indxd = 2;   {indexed}
		immed = 3;   {immediate}
		ireg = 4;   {indirect register}
		imem = 5;   {indirect memory}
		iindxd = 6;   {indirect indexed}
		pcrel = 7;   {Pc relative}

    {opcodes}
		INDENT = 0;
		IA = 1;
		IS = 2;
		IM = 3;
		ID = 4;
		FN = 5;
		FA = 6;
		FS = 7;
		FM = 8;
		FD = 9;
		BI = 10;
		BO = 11;
		BA = 12;
		IC = 13;
		FC = 14;
		JSR = 15;
		BKT = 16;
		LD = 17;
		STO = 18;
		LDA = 19;
		FLT = 20;
		FIX = 21;
		J = 22;
		SR = 23;
		SL = 24;
		RD = 25;
		WR = 26;
		TRNG = 27;
		HALT = 31;
		maxSmallInt = 16383;
		minSmallInt = -16383;

	TYPE
		regaddr = 0..15;
		memaddr = 0..maxSmallInt;

		word = SET OF 0..15;

		bytee = SET OF 0..7;

		doubleword = SET OF 0..31;

		wordrange = minSmallInt..maxSmallInt;

		byterange = -128..127;

		icrec = RECORD
				CASE boolean OF
					true: (
							wf: word
					);
					false: (
							sintf: wordrange
					)
			END;

		bcrec = RECORD
				CASE boolean OF
					true: (
							bytef: bytee
					);
					false: (
							hintf: byterange
					)
			END;

		rcrec = RECORD
				CASE boolean OF
					true: (
							rf: real
					);
					false: (
							b1, b4, b5, b6: bytee
					)
			END;
		ccrec = RECORD
				CASE boolean OF
					true: (
							c: PACKED ARRAY[1..2] OF char
					);
					false: (
							wf: word
					)
			END;
	VAR
		Mem: ARRAY[memaddr] OF bytee; {Memory}
		Regs: ARRAY[regaddr] OF word; {Registers}
		Pc: memaddr;  {Program Counter}
		Ir: word;   {Instruction Register}
		StopProgram: boolean;  {Machine halt flag}
		Lt, Eq, Gt: boolean;
		oldPc: memaddr;
		Icr: icrec;
		Rcr: rcrec;
		Ccr: ccrec;
		Bcr: bcrec;
		Arg, Memfname, Dumpfname: PACKED ARRAY[1..20] OF char;
		Dumpf: FILE OF byterange;
		outFile, inFile: text;
		MemImage: boolean;  {true iff there is a nonempty memory image}
		Trace, Dump: boolean;
		i: integer;
		RegIndex: integer; { used to zero registers }
		MemoryFileName, ListingFileName: STRING;
		VRefNum: integer;
		finderInfo: FInfo;
		noError: OSErr;

	PROCEDURE clearLine (n: integer);
		VAR
			r: rect;
	BEGIN
	r.top := n - 15;
	r.left := 1;
	r.bottom := n + 15;
	r.right := 396;
	eraseRect(r);
	END;

	FUNCTION OldFileName (p: STRING): STRING;
		VAR
			topLeft: point;
			TypeList: SFTypeList;
			reply: SFReply;
	BEGIN
	topleft.v := 100;
	topLeft.h := 100;
	typelist[0] := 'TEXT';
	OldFileName := '';
	sfGetFile(topLeft, '', NIL, 1, typeList, NIL, reply);
	IF reply.good THEN BEGIN
		OldFileName := (reply.fName);
		vRefNum := reply.vRefNum;
		END;
	END;

	FUNCTION NewFileName (p: STRING): STRING;
		VAR
			topLeft: point;
			reply: SFReply;
			NFN: STRING;
	BEGIN
	topleft.v := 100;
	topLeft.h := 100;
	NFN := 'Output';
	NewFileName := NFN;
	sfPutFile(topLeft, '', NFN, NIL, reply);
	IF reply.good THEN BEGIN
		NewFileName := (reply.fName);
		vRefNum := reply.vRefNum;
		END;
	END;

	PROCEDURE convertWtoExternal (VAR Rcr: rcrec);
		VAR
			i: integer;
	BEGIN

	END;

	PROCEDURE convertWtoInternal (VAR Rcr: rcrec);
		VAR
			i: integer;
	BEGIN

	END;

	PROCEDURE wordtobytes (w: word; VAR hibytee, lobytee: bytee);
		VAR
			i: integer;
	BEGIN
	lobytee := [];
	hibytee := [];
	FOR i := 0 TO 7 DO BEGIN
		IF i IN w THEN
			lobytee := lobytee + [i];
		IF i + 8 IN w THEN
			hibytee := hibytee + [i]
		END {for}
	END; {procedure wordtobytes}


	PROCEDURE bytestoword (hibytee, lobytee: bytee; VAR w: word);
		VAR
			i: integer;
	BEGIN
	w := [];
	FOR i := 0 TO 7 DO BEGIN
		IF i IN lobytee THEN
			w := w + [i];
		IF i IN hibytee THEN
			w := w + [i + 8]
		END {for}
	END; {procedure bytestoword}

	PROCEDURE wordstodoubleword (loword, hiword: word; VAR dw: doubleword);
		VAR
			i: integer;
	BEGIN
	dw := [];
	FOR i := 0 TO 15 DO BEGIN
		IF i IN loword THEN
			dw := dw + [i];
		IF i IN hiword THEN
			dw := dw + [i + 16]
		END {for}
	END; {procedure wordstodoubleword}

	PROCEDURE doublewordtowords (dw: doubleword; VAR loword, hiword: word);
		VAR
			i: integer;
	BEGIN
	loword := [];
	hiword := [];
	FOR i := 0 TO 15 DO BEGIN
		IF i IN dw THEN
			loword := loword + [i];
		IF i + 16 IN dw THEN
			hiword := hiword + [i]
		END {for}
	END; {procedure doublewordtowords}

	PROCEDURE writeinst (loc, opcode, field1, field2, field3, addr2: integer);
    (*----------------------------------------------*}
{    * writeinst -                                   *}
{    *     writes out an instruction to the listing  *}
{    *-----------------------------------------------*)

		PROCEDURE showreg (reg: integer);
          (* write a register symbol to the listing *)
		BEGIN
		write(outFile, 'r', reg : 1)
		END; (* showreg *)

		PROCEDURE showaddr (location, mode, r2, a2: integer);
          (* write a symbolic general address to the listing *)

			PROCEDURE showindxd (r, a: integer);
             (* write a symbolic indexed address to the listing *)
			BEGIN
			write(outFile, a : 1, '(');
			showreg(r);
			write(outFile, ')')
			END; (* showindxd *)

		BEGIN (* showaddr *)
		CASE mode OF
		dreg: 
			showreg(r2);
		dmem: 
			write(outFile, a2 : 1);
		indxd: 
			showindxd(r2, a2);
		immed: 
			write(outFile, '#', a2 : 1);
		ireg:  BEGIN
			write(outFile, '*');
			showreg(r2)
			END;
		imem: 
			write(outFile, '*', a2 : 1);
		iindxd:  BEGIN
			write(outFile, '*');
			showindxd(r2, a2)
			END;
		pcrel: 
			write(outFile, '&', a2 : 1);
		OTHERWISE
			write(outFile, 'invalid addressing mode = ', mode : 1);
		END (* case *)
		END; (* showaddr *)

	BEGIN (* writeinst *)
	write(outFile, ' ' : 5, loc : 5, ' ' : 5);
	CASE opcode OF
	INDENT: 
		write(outFile, 'in    ');
	IA: 
		write(outFile, 'ia    ');
	IS: 
		write(outFile, 'is    ');
	IM: 
		write(outFile, 'im    ');
	ID: 
		write(outFile, 'id    ');

	FN: 
		write(outFile, 'fn    ');
	FA: 
		write(outFile, 'fa    ');
	FS: 
		write(outFile, 'fs    ');
	FM: 
		write(outFile, 'fm    ');
	FD: 
		write(outFile, 'fd    ');

	BI: 
		write(outFile, 'bi    ');
	BO: 
		write(outFile, 'bo    ');
	BA: 
		write(outFile, 'ba    ');

	IC: 
		write(outFile, 'ic    ');
	FC: 
		write(outFile, 'fc    ');

	JSR: 
		write(outFile, 'jsr   ');

	BKT: 
		write(outFile, 'bkt   ');
	LD: 
		write(outFile, 'ld    ');
	STO: 
		write(outFile, 'sto   ');
	LDA: 
		write(outFile, 'lda   ');

	FLT: 
		write(outFile, 'flt   ');
	FIX: 
		write(outFile, 'fix   ');

	J: 
		CASE field1 OF
		0: 
			write(outFile, 'jmp   ');
		1: 
			write(outFile, 'jlt   ');
		2: 
			write(outFile, 'jle   ');
		3: 
			write(outFile, 'jeq   ');
		4: 
			write(outFile, 'jne   ');
		5: 
			write(outFile, 'jge   ');
		6: 
			write(outFile, 'jgt   ');
		7: 
			write(outFile, 'nop   ');
		OTHERWISE
			write(outFile, 'invalid jump instruction, type = ', field1 : 1, ': ');
		END; (* case *)

	SR: 
		CASE field2 OF
		0: 
			write(outFile, 'srz   ');
		1: 
			write(outFile, 'sro   ');
		2: 
			write(outFile, 'sre   ');
		3: 
			write(outFile, 'src   ');
		4: 
			write(outFile, 'srcz  ');
		5: 
			write(outFile, 'srco  ');
		6: 
			write(outFile, 'srce  ');
		7: 
			write(outFile, 'srcc  ');
		OTHERWISE
			write(outFile, 'invalid sr instruction, mode = ', field2 : 1, ': ');
		END; (* case *)

	SL: 
		CASE field2 OF
		0: 
			write(outFile, 'slz   ');
		1: 
			write(outFile, 'slo   ');
		2: 
			write(outFile, 'sle   ');
		3: 
			write(outFile, 'slc   ');
		4: 
			write(outFile, 'slcz  ');
		5: 
			write(outFile, 'slco  ');
		6: 
			write(outFile, 'slce  ');
		7: 
			write(outFile, 'slcc  ');
		OTHERWISE
			write(outFile, 'invalid sl instruction, mode = ', field2 : 1, ': ');
		END; (* case *)

	RD: 
		CASE field1 OF
		0: 
			write(outFile, 'rdi   ');
		1: 
			write(outFile, 'rdf   ');
		2: 
			write(outFile, 'rdbd  ');
		3: 
			write(outFile, 'rdbw  ');
		4: 
			write(outFile, 'rdod  ');
		5: 
			write(outFile, 'rdow  ');
		6: 
			write(outFile, 'rdhd  ');
		7: 
			write(outFile, 'rdhw  ');
		8: 
			write(outFile, 'rdch  ');
		9: 
			write(outFile, 'rdst  ');
		10: 
			write(outFile, 'rdin  ');
		11: 
			write(outFile, 'rdln  ');
		OTHERWISE
			write(outFile, 'invalid rd instruction, fmt = ', field1 : 1, ': ');
		END; (* case *)

	WR: 
		CASE field1 OF
		0: 
			write(outFile, 'wri   ');
		1: 
			write(outFile, 'wrf   ');
		2: 
			write(outFile, 'wrbd  ');
		3: 
			write(outFile, 'wrbw  ');
		4: 
			write(outFile, 'wrod  ');
		5: 
			write(outFile, 'wrow  ');
		6: 
			write(outFile, 'wrhd  ');
		7: 
			write(outFile, 'wrhw  ');
		8: 
			write(outFile, 'wrch  ');
		9: 
			write(outFile, 'wrst  ');
		11: 
			write(outFile, 'wrnl  ');
		OTHERWISE
			write(outFile, 'invalid wr instruction, fmt = ', field1 : 1, ': ');
		END; (* case *)

	TRNG: 
		write(outFile, 'trng  ');
	HALT: 
		write(outFile, 'halt  ');

	OTHERWISE
		write(outFile, 'invalid instruction, opcode = ', opcode : 1, ': ');
	END; (* case *)

	IF opcode IN [INDENT..FLT, TRNG] THEN         (* 2 address *)
		BEGIN
		showreg(field1);
		write(outFile, ', ');
		showaddr(loc, field2, field3, addr2)
		END (* if *)
	ELSE IF ((opcode = J) AND (field1 <> 7)) OR ((opcode = WR) AND (field1 <> 11)) OR (opcode = RD) THEN   (* jump but not nop *)
 (* write but not wrnl *)
		showaddr(loc, field2, field3, addr2)
	ELSE IF opcode IN [SR, SL] THEN       (* 1 register address *)
		BEGIN
		showreg(field1);
		write(outFile, ', ');
		write(outFile, field3)  (* amnt *)
		END; (* else if *)

	writeln(outFile)
	END; (* writeinst *)


	PROCEDURE writeboolean (b: boolean);
	BEGIN
	IF b THEN
		write(outFile, 'TRUE')
	ELSE
		write(outFile, 'FALSE')
	END; (* writeboolean *)

	PROCEDURE displayregs;
		VAR
			i: integer;
	BEGIN
	writeln(outFile, '########################################');
	write(outFile, '  Pc = ', Pc : 1, '  Lt = ');
	writeboolean(Lt);
	write(outFile, '  Eq = ');
	writeboolean(Eq);
	write(outFile, '  Gt = ');
	writeboolean(Gt);
	writeln(outFile);
	FOR i := 0 TO REGSIZ - 1 DO BEGIN
		Icr.wf := Regs[i];
		write(outFile, '  Reg[', i : 2, '] = ', Icr.sintf : 5);
		IF (i + 1) MOD 4 = 0 THEN
			writeln(outFile)
		END; {for}
	writeln(outFile, '########################################')
	END; (* displayregs *)


	PROCEDURE fetchword (VAR w: word);
	BEGIN
	bytestoword(Mem[Pc], Mem[Pc + 1], w);
	Pc := (Pc + 2) MOD MEMSIZ
	END; {procedure fetchword}


	FUNCTION eac (mode: integer; reg: regaddr; VAR w2: integer): integer;
    {perform effective address calculation -- update Pc as needed for 2}
{    word instructions.}
		VAR
			taddr: integer;    {temporary address}
	BEGIN
	w2 := 0;
	CASE mode OF
	dreg: 
		taddr := -1;			{indicates direct reg}
	dmem:  BEGIN
		fetchword(Icr.wf);
       			{Icr.wf := Icr.wf - [15];}
		taddr := Icr.sintf;
		w2 := taddr
		END;
	indxd:  BEGIN
		fetchword(Icr.wf);
		w2 := Icr.sintf;
		Icr.wf := Regs[reg];
		taddr := (w2 + Icr.sintf) MOD MEMSIZ
		END;
	immed:  BEGIN
		taddr := Pc;
		bytestoword(Mem[taddr], Mem[taddr + 1], Icr.wf);
		w2 := Icr.sintf;
		Pc := (Pc + 2) MOD MEMSIZ
		END;
	ireg:  BEGIN
		Icr.wf := Regs[reg]; {- [15]}
		taddr := Icr.sintf
		END;
	imem:  BEGIN
		fetchword(Icr.wf);
				   {Icr.wf := Icr.wf - [15];}
		w2 := Icr.sintf;
		bytestoword(Mem[w2], Mem[w2 + 1], Icr.wf);
				   {Icr.wf := Icr.wf - [15];}
		taddr := Icr.sintf
		END;
	iindxd:  BEGIN
		fetchword(Icr.wf);
		w2 := Icr.sintf;
		Icr.wf := Regs[reg];
		taddr := (w2 + Icr.sintf) MOD MEMSIZ;
		bytestoword(Mem[taddr], Mem[taddr + 1], Icr.wf);
				   {Icr.wf := Icr.wf - [15];}
		taddr := Icr.sintf
		END;
	pcrel:  BEGIN
		fetchword(Icr.wf);
		w2 := Icr.sintf;
		taddr := (w2 + Pc) MOD MEMSIZ
		END
	END; {case of mode}
	eac := taddr
	END; {function eac}


	PROCEDURE twoaddr (opcode, r1, amode, r2: integer; VAR w2: integer);
    {perform two address operation}
		VAR
			addr: integer;
			wd: word;
			iop1, iop2, ans: integer;
			rop1, rop2: real;
			i: integer;
	BEGIN
	addr := eac(amode, r2, w2);
	IF opcode IN [INDENT, IA, IS, IM, ID, IC] THEN BEGIN
		IF addr >= 0 THEN
			bytestoword(Mem[addr], Mem[addr + 1], Icr.wf)
		ELSE
			Icr.wf := Regs[r2];
		iop2 := Icr.sintf;
		Icr.wf := Regs[r1];
		iop1 := Icr.sintf;
		CASE opcode OF
		INDENT: 
			ans := -iop2;
		IA: 
			ans := iop1 + iop2;
		IS: 
			ans := iop1 - iop2;
		IM: 
			ans := iop1 * iop2;
		ID: 
			ans := iop1 DIV iop2;
		IC:  BEGIN
			ans := iop1;    {to restore state of Regs[r1]}
			Lt := false;
			Eq := false;
			Gt := false;
			IF iop1 < iop2 THEN
				Lt := true
			ELSE IF iop1 = iop2 THEN
				Eq := true
			ELSE
				Gt := true
			END
		END; {case of opcode}
		Icr.sintf := ans;
		Regs[r1] := Icr.wf
		END
	ELSE IF opcode IN [FN, FA, FS, FM, FD, FC] THEN BEGIN
		wordtobytes(Regs[r1], Rcr.b1, Rcr.b4);
		wordtobytes(Regs[(r1 + 1) MOD REGSIZ], Rcr.b5, Rcr.b6);
		convertWtoExternal(Rcr);
		rop1 := Rcr.rf;
		IF addr >= 0 THEN BEGIN
{    Rcr.b2 := [];}
{    Rcr.b3 := [];}
			Rcr.b1 := Mem[addr];
			Rcr.b4 := Mem[addr + 1];
			Rcr.b5 := Mem[addr + 2];
			Rcr.b6 := Mem[addr + 3];
			END
		ELSE BEGIN
			wordtobytes(Regs[r2], Rcr.b1, Rcr.b4);
			wordtobytes(Regs[(r2 + 1) MOD REGSIZ], Rcr.b5, Rcr.b6);
			END;
		convertWtoExternal(Rcr);
		rop2 := Rcr.rf;
		CASE opcode OF
		FN: 
			rop1 := -rop2;
		FA: 
			rop1 := rop1 + rop2;
		FS: 
			rop1 := rop1 - rop2;
		FM: 
			rop1 := rop1 * rop2;
		FD: 
			rop1 := rop1 / rop2;
		FC:  BEGIN
			Lt := false;
			Eq := false;
			Gt := false;
			IF rop1 < rop2 THEN
				Lt := true
			ELSE IF rop1 = rop2 THEN
				Eq := true
			ELSE
				Gt := true
			END
		END; {case of opcode}
		Rcr.rf := rop1;
		convertWtoInternal(Rcr);
		bytestoword(Rcr.b1, Rcr.b4, Regs[r1]);
		bytestoword(Rcr.b5, Rcr.b6, Regs[(r1 + 1) MOD REGSIZ])
		END
	ELSE IF opcode IN [BI, BO, BA] THEN BEGIN
		IF addr >= 0 THEN
			bytestoword(Mem[addr], Mem[addr + 1], wd)
		ELSE
			wd := Regs[r2];
		CASE opcode OF
		BI:  BEGIN
			Regs[r1] := [];
			FOR i := 0 TO 15 DO
				IF NOT (i IN wd) THEN
					Regs[r1] := Regs[r1] + [i]
			END;
		BO: 
			Regs[r1] := Regs[r1] + wd;
		BA: 
			Regs[r1] := Regs[r1] * wd
		END {case of opcode}
		END
	ELSE IF opcode = JSR THEN BEGIN
		Icr.sintf := Pc;
		Regs[r1] := Icr.wf;
		IF addr >= 0 THEN
			Pc := addr
		ELSE BEGIN
			writeln(outFile, 'JSR to a Register ', Pc - 2 : 1);
			StopProgram := true
			END
		END
	ELSE IF opcode = BKT THEN
		IF addr < 0 THEN BEGIN
			writeln(outFile, 'Address of BKT is a Register ', Pc - 2 : 1);
			StopProgram := true
			END
		ELSE BEGIN
			Icr.wf := Regs[r1];
			iop2 := Icr.sintf;
			Icr.wf := Regs[(r1 + 1) MOD REGSIZ];
			FOR i := 0 TO Icr.sintf - 1 DO
				Mem[(addr + i) MOD MEMSIZ] := Mem[(iop2 + i) MOD MEMSIZ]
			END {if}
	ELSE IF opcode = LD THEN
		IF addr >= 0 THEN
			bytestoword(Mem[addr], Mem[addr + 1], Regs[r1])
		ELSE
			Regs[r1] := Regs[r2]
	ELSE IF opcode = STO THEN
		IF addr >= 0 THEN BEGIN
			wordtobytes(Regs[r1], Mem[addr], Mem[addr + 1]);
			IF Trace THEN BEGIN
				bytestoword(Mem[addr], Mem[addr + 1], Icr.wf);
				writeln(outFile, '***** value ', Icr.sintf : 1, ' stored at location ', addr : 1)
				END
			END
		ELSE
			Regs[r2] := Regs[r1]
	ELSE IF opcode = LDA THEN
		IF addr >= 0 THEN BEGIN
			Icr.sintf := addr;
			Regs[r1] := Icr.wf
			END
		ELSE BEGIN
			writeln(outFile, 'Address of LDA is a Register ', Pc - 2 : 1);
			StopProgram := true
			END
	ELSE IF opcode = FLT THEN BEGIN
		IF addr >= 0 THEN
			bytestoword(Mem[addr], Mem[addr + 1], Icr.wf)
		ELSE
			Icr.wf := Regs[r2];
		Rcr.rf := Icr.sintf;
		convertWtoInternal(Rcr);
		bytestoword(Rcr.b1, Rcr.b4, Regs[r1]);
		bytestoword(Rcr.b5, Rcr.b6, Regs[(r1 + 1) MOD REGSIZ])
		END
	ELSE IF opcode = FIX THEN BEGIN
		IF addr >= 0 THEN BEGIN
{    Rcr.b2 := [];}
{    Rcr.b3 := [];}
			Rcr.b1 := Mem[addr];
			Rcr.b4 := Mem[addr + 1];
			Rcr.b5 := Mem[addr + 2];
			Rcr.b6 := Mem[addr + 3];
			END
		ELSE BEGIN
			wordtobytes(Regs[r2], Rcr.b1, Rcr.b4);
			wordtobytes(Regs[(r2 + 1) MOD REGSIZ], Rcr.b5, Rcr.b6);
			END; {if}
		convertWtoExternal(Rcr);
		Icr.sintf := round(Rcr.rf);
		Regs[r1] := Icr.wf
		END
	ELSE IF opcode = TRNG THEN
		IF addr >= 0 THEN BEGIN
			bytestoword(Mem[addr], Mem[addr + 1], Icr.wf);
			iop1 := Icr.sintf;		{lower bound}
			bytestoword(Mem[addr + 2], Mem[addr + 3], Icr.wf);
			iop2 := Icr.sintf;		{upper bound}
			Icr.wf := Regs[r1];
			ans := Icr.sintf;		{test expression}
{ ***** On next line, Icr.sintf should be ans?! ***** }
			IF (ans < iop1) OR (ans > iop2) THEN BEGIN
				writeln(outFile, 'Index expression out of bounds at ', Pc - 2 : 1, ' (expr = ', ans : 1, ' lower = ', iop1 : 1, ' upper = ', iop2 : 1, ')');
				StopProgram := true
				END
			END
		ELSE BEGIN
			writeln(outFile, 'Address of TRNG is a register at ', Pc - 2 : 1);
			StopProgram := true
			END
	END; {procedure twoaddr}


	PROCEDURE jop (jmode, amode, reg: integer; VAR w2: integer);
    {perform jump operation}
		VAR
			addr: integer;
	BEGIN
	addr := eac(amode, reg, w2);
	IF addr < 0 THEN BEGIN
		writeln(outFile, 'Jump to a Register ', Pc - 2 : 1);
		StopProgram := true
		END
	ELSE IF jmode IN [0, 1, 2, 3, 4, 5, 6, 7] THEN
		CASE jmode OF
		0: 
			Pc := addr;
		1: 
			IF Lt THEN
				Pc := addr;
		2: 
			IF Lt OR Eq THEN
				Pc := addr;
		3: 
			IF Eq THEN
				Pc := addr;
		4: 
			IF NOT Eq THEN
				Pc := addr;
		5: 
			IF Gt OR Eq THEN
				Pc := addr;
		6: 
			IF Gt THEN
				Pc := addr;
		7: 
			;
		    {null}
		END {case of jmode}
	ELSE BEGIN
		writeln(outFile, 'Invalid Jump Mode ', Pc - 2 : 1);
		StopProgram := true
		END {if}
	END; {procedure jop}


	PROCEDURE srop (reg, smode, amnt: integer);
    {perform right shift}
		VAR
			w1, w2, w3: word;
			i, j: integer;
			flag: boolean;
	BEGIN
	CASE smode OF
	0:  BEGIN
		w1 := Regs[reg];
		IF amnt = 0 THEN
			w1 := []
		ELSE
			FOR i := 1 TO amnt DO BEGIN
				FOR j := 0 TO 15 - i DO BEGIN
					w1 := w1 - [j];
					IF j + 1 IN w1 THEN
						w1 := w1 + [j]
					END; {for j}
				w1 := w1 - [16 - i]
				END; {for i}
		Regs[reg] := w1
		END;
	1:  BEGIN
		w1 := Regs[reg];
		IF amnt = 0 THEN
			w1 := [0..15]
		ELSE
			FOR i := 1 TO amnt DO BEGIN
				FOR j := 0 TO 15 - i DO BEGIN
					w1 := w1 - [j];
					IF j + 1 IN w1 THEN
						w1 := w1 + [j]
					END; {for j}
				w1 := w1 + [16 - i]
				END; {for i}
		Regs[reg] := w1
		END;
	2:  BEGIN
		w1 := Regs[reg];
		IF amnt = 0 THEN
			IF 15 IN w1 THEN
				w1 := [0..15]
			ELSE
				w1 := []
		ELSE
			FOR i := 1 TO amnt DO
				FOR j := 0 TO 14 DO BEGIN
					w1 := w1 - [j];
					IF j + 1 IN w1 THEN
						w1 := w1 + [j]
					END; {for j}
		Regs[reg] := w1
		END;
	3:  BEGIN				{SRC}
		w1 := Regs[reg];
		IF amnt <> 0 THEN
			FOR i := 1 TO amnt DO BEGIN
				flag := 0 IN w1;
				FOR j := 0 TO 14 DO BEGIN
					w1 := w1 - [j];
					IF j + 1 IN w1 THEN
						w1 := w1 + [j]
					END; {for j}
				IF flag THEN
					w1 := w1 + [15]
				ELSE
					w1 := w1 - [15]
				END; {for i}
		Regs[reg] := w1
		END;
	4:  BEGIN				{SRCZ}
		w1 := Regs[reg];
		w2 := Regs[(reg + 1) MOD REGSIZ];
		IF amnt = 0 THEN BEGIN
			w2 := w1;
			w1 := []
			END
		ELSE BEGIN
			FOR i := 1 TO amnt DO BEGIN
				FOR j := 0 TO 14 DO BEGIN
					w2 := w2 - [j];
					IF j + 1 IN w2 THEN
						w2 := w2 + [j]
					END; {for j}
				IF 0 IN w1 THEN
					w2 := w2 + [15]
				ELSE
					w2 := w2 - [15];
				FOR j := 0 TO 15 - i DO BEGIN
					w1 := w1 - [j];
					IF j + 1 IN w1 THEN
						w1 := w1 + [j]
					END; {for j}
				w1 := w1 - [16 - i]
				END
			END; {for i}
 {else}
		Regs[reg] := w1;
		Regs[(reg + 1) MOD REGSIZ] := w2
		END;
	5:  BEGIN				{SRCO}
		w1 := Regs[reg];
		w2 := Regs[(reg + 1) MOD REGSIZ];
		IF amnt = 0 THEN BEGIN
			w2 := w1;
			w1 := [0..15]
			END
		ELSE BEGIN
			FOR i := 1 TO amnt DO BEGIN
				FOR j := 0 TO 14 DO BEGIN
					w2 := w2 - [j];
					IF j + 1 IN w2 THEN
						w2 := w2 + [j]
					END; {for j}
				IF 0 IN w1 THEN
					w2 := w2 + [15]
				ELSE
					w2 := w2 - [15];
				FOR j := 0 TO 15 - i DO BEGIN
					w1 := w1 - [j];
					IF j + 1 IN w1 THEN
						w1 := w1 + [j]
					END; {for j}
				w1 := w1 + [16 - i]
				END
			END; {for i}
 {else}
		Regs[reg] := w1;
		Regs[(reg + 1) MOD REGSIZ] := w2
		END;
	6:  BEGIN
		w1 := Regs[reg];
		w2 := Regs[(reg + 1) MOD REGSIZ];
		IF amnt = 0 THEN BEGIN
			w2 := w1;
			IF 15 IN w2 THEN
				w1 := [0..15]
			ELSE
				w1 := []
			END
		ELSE BEGIN
			FOR i := 1 TO amnt DO BEGIN
				FOR j := 0 TO 14 DO BEGIN
					w2 := w2 - [j];
					IF j + 1 IN w2 THEN
						w2 := w2 + [j]
					END; {for j}
				IF 0 IN w1 THEN
					w2 := w2 + [15]
				ELSE
					w2 := w2 - [15];
				FOR j := 0 TO 15 - i DO BEGIN
					w1 := w1 - [j];
					IF j + 1 IN w1 THEN
						w1 := w1 + [j]
					END
				END {for j}
			END; {for i}
 {else}
		Regs[reg] := w1;
		Regs[(reg + 1) MOD REGSIZ] := w2
		END;
	7:  BEGIN
		w1 := Regs[reg];
		w2 := Regs[(reg + 1) MOD REGSIZ];
		IF amnt = 0 THEN BEGIN
			w3 := w2;
			w2 := w1;
			w1 := w3
			END
		ELSE BEGIN
			FOR i := 1 TO amnt DO BEGIN
				flag := 0 IN w2;
				FOR j := 0 TO 14 DO BEGIN
					w2 := w2 - [j];
					IF j + 1 IN w2 THEN
						w2 := w2 + [j]
					END; {for j}
				IF 0 IN w1 THEN
					w2 := w2 + [15]
				ELSE
					w2 := w2 - [15];
				FOR j := 0 TO 14 DO BEGIN
					w1 := w1 - [j];
					IF j + 1 IN w1 THEN
						w1 := w1 + [j]
					END; {for j}
				IF flag THEN
					w1 := w1 + [15]
				ELSE
					w1 := w1 - [15]
				END
			END; {for i}
 {else}
		Regs[reg] := w1;
		Regs[(reg + 1) MOD REGSIZ] := w2
		END
	END {case of smode}
	END; {procedure srop}


	PROCEDURE slop (reg, smode, amnt: integer);
    {perform left shift}
		VAR
			w1, w2, w3: word;
			i, j: integer;
			flag: boolean;
	BEGIN
	CASE smode OF
	0:  BEGIN
		w1 := Regs[reg];
		IF amnt = 0 THEN
			w1 := []
		ELSE
			FOR i := 1 TO amnt DO BEGIN
				FOR j := 14 DOWNTO i - 1 DO BEGIN
					w1 := w1 - [j + 1];
					IF j IN w1 THEN
						w1 := w1 + [j + 1]
					END; {for j}
				w1 := w1 - [i - 1]
				END; {for i}
		Regs[reg] := w1
		END;
	1:  BEGIN
		w1 := Regs[reg];
		IF amnt = 0 THEN
			w1 := [0..15]
		ELSE
			FOR i := 1 TO amnt DO BEGIN
				FOR j := 14 DOWNTO i - 1 DO BEGIN
					w1 := w1 - [j + 1];
					IF j IN w1 THEN
						w1 := w1 + [j + 1]
					END; {for j}
				w1 := w1 + [i - 1]
				END; {for i}
		Regs[reg] := w1
		END;
	2:  BEGIN
		w1 := Regs[reg];
		IF amnt = 0 THEN
			IF 0 IN w1 THEN
				w1 := [0..15]
			ELSE
				w1 := []
		ELSE
			FOR i := 1 TO amnt DO
				FOR j := 14 DOWNTO i - 1 DO BEGIN
					w1 := w1 - [j + 1];
					IF j IN w1 THEN
						w1 := w1 + [j + 1]
					END; {for j}
		Regs[reg] := w1
		END;
	3:  BEGIN
		w1 := Regs[reg];
		IF amnt <> 0 THEN
			FOR i := 1 TO amnt DO BEGIN
				flag := 15 IN w1;
				FOR j := 14 DOWNTO 0 DO BEGIN
					w1 := w1 - [j + 1];
					IF j IN w1 THEN
						w1 := w1 + [j + 1]
					END; {for j}
				IF flag THEN
					w1 := w1 + [0]
				ELSE
					w1 := w1 - [0]
				END; {for i}
		Regs[reg] := w1
		END;
	4:  BEGIN
		w1 := Regs[reg];
		w2 := Regs[(reg + 1) MOD REGSIZ];
		IF amnt = 0 THEN BEGIN
			w1 := w2;
			w2 := []
			END
		ELSE BEGIN
			FOR i := 1 TO amnt DO BEGIN
				FOR j := 14 DOWNTO 0 DO BEGIN
					w1 := w1 - [j + 1];
					IF j IN w1 THEN
						w1 := w1 + [j + 1]
					END; {for j}
				IF 15 IN w2 THEN
					w1 := w1 + [0]
				ELSE
					w1 := w1 - [0];
				FOR j := 14 DOWNTO i - 1 DO BEGIN
					w2 := w2 - [j + 1];
					IF j IN w2 THEN
						w2 := w2 + [j + 1]
					END; {for j}
				w2 := w2 - [i - 1]
				END
			END; {for i}
 {else}
		Regs[reg] := w1;
		Regs[(reg + 1) MOD REGSIZ] := w2
		END;
	5:  BEGIN
		w1 := Regs[reg];
		w2 := Regs[(reg + 1) MOD REGSIZ];
		IF amnt = 0 THEN BEGIN
			w1 := w2;
			w2 := [0..15]
			END
		ELSE BEGIN
			FOR i := 1 TO amnt DO BEGIN
				FOR j := 14 DOWNTO 0 DO BEGIN
					w1 := w1 - [j + 1];
					IF j IN w1 THEN
						w1 := w1 + [j + 1]
					END; {for j}
				IF 15 IN w2 THEN
					w1 := w1 + [0]
				ELSE
					w1 := w1 - [0];
				FOR j := 14 DOWNTO i - 1 DO BEGIN
					w2 := w2 - [j + 1];
					IF j IN w2 THEN
						w2 := w2 + [j + 1]
					END; {for j}
				w2 := w2 + [i - 1]
				END
			END; {for i}
 {else}
		Regs[reg] := w1;
		Regs[(reg + 1) MOD REGSIZ] := w2
		END;
	6:  BEGIN
		w1 := Regs[reg];
		w2 := Regs[(reg + 1) MOD REGSIZ];
		IF amnt = 0 THEN BEGIN
			w1 := w2;
			IF 0 IN w2 THEN
				w2 := [0..15]
			ELSE
				w2 := []
			END
		ELSE BEGIN
			FOR i := 1 TO amnt DO BEGIN
				FOR j := 14 DOWNTO 0 DO BEGIN
					w1 := w1 - [j + 1];
					IF j IN w1 THEN
						w1 := w1 + [j + 1]
					END; {for j}
				IF 15 IN w2 THEN
					w1 := w1 + [0]
				ELSE
					w1 := w1 - [0];
				FOR j := 14 DOWNTO i - 1 DO BEGIN
					w2 := w2 - [j + 1];
					IF j IN w2 THEN
						w2 := w2 + [j + 1]
					END
				END {for j}
			END; {for i}
 {else}
		Regs[reg] := w1;
		Regs[(reg + 1) MOD REGSIZ] := w2
		END;
	7:  BEGIN
		w1 := Regs[reg];
		w2 := Regs[(reg + 1) MOD REGSIZ];
		IF amnt = 0 THEN BEGIN
			w3 := w1;
			w1 := w2;
			w2 := w3
			END
		ELSE BEGIN
			FOR i := 1 TO amnt DO BEGIN
				flag := 15 IN w1;
				FOR j := 14 DOWNTO 0 DO BEGIN
					w1 := w1 - [j + 1];
					IF j IN w1 THEN
						w1 := w1 + [j + 1]
					END; {for j}
				IF 15 IN w2 THEN
					w1 := w1 + [0]
				ELSE
					w1 := w1 - [0];
				FOR j := 14 DOWNTO 0 DO BEGIN
					w2 := w2 - [j + 1];
					IF j IN w2 THEN
						w2 := w2 + [j + 1]
					END; {for j}
				IF flag THEN
					w2 := w2 + [0]
				ELSE
					w2 := w2 - [0]
				END
			END; {for i}
 {else}
		Regs[reg] := w1;
		Regs[(reg + 1) MOD REGSIZ] := w2
		END
	END {case of smode}
	END; {procedure slop}


	PROCEDURE rdop (rmode, amode, reg: integer; VAR w2: integer);
    {perform a read operation}
		VAR
			wd: word;
			flag: boolean;
			addr: integer;
			i: integer;
			ival: integer;
			ch: char;
	BEGIN
	IF rmode IN [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11] THEN BEGIN
		addr := eac(amode, reg, w2);
		CASE rmode OF
		0:  BEGIN			{RDI--read integer}
			read(inFile, i);
			IF (i > maxSmallInt) OR (i < -maxSmallInt) THEN BEGIN
				writeln(outFile, 'Invalid Input: RDI ', Pc - 2 : 1);
				StopProgram := true
				END
			ELSE BEGIN
				Icr.sintf := i;
				IF addr >= 0 THEN
					wordtobytes(Icr.wf, Mem[addr], Mem[addr + 1])
				ELSE
					Regs[reg] := Icr.wf
				END {else}
			END;
		1:  BEGIN			{RDF--read floating point}
			read(inFile, Rcr.rf);
			convertWtoInternal(Rcr);
			IF addr >= 0 THEN BEGIN
				Mem[addr] := Rcr.b1;
				Mem[addr + 1] := Rcr.b4;
				Mem[addr + 2] := Rcr.b5;
				Mem[addr + 3] := Rcr.b6
				END
			ELSE BEGIN
				bytestoword(Rcr.b1, Rcr.b4, Regs[reg]);
				bytestoword(Rcr.b5, Rcr.b6, Regs[(reg + 1) MOD REGSIZ])
				END {else}
			END;
		2:  BEGIN			{RDBD--read binary digit}
			read(inFile, ch);
			IF ch = '0' THEN
				IF addr >= 0 THEN
					Mem[addr] := Mem[addr] - [7]
				ELSE
					Regs[addr] := Regs[addr] - [15]
			ELSE IF ch = '1' THEN
				IF addr >= 0 THEN
					Mem[addr] := Mem[addr] + [7]
				ELSE
					Regs[addr] := Regs[addr] + [15]
			ELSE BEGIN
				writeln(outFile, 'Invalid Input: RDBD ', Pc - 2 : 1);
				StopProgram := true
				END
			END;
		3:  BEGIN			{RDBW--read binary word}
			i := 15;
			flag := true;
			WHILE (i >= 0) AND flag DO BEGIN
				read(inFile, ch);
				IF ch = '1' THEN
					wd := wd + [i]
				ELSE IF ch = '0' THEN
					wd := wd - [i]
				ELSE BEGIN
					writeln(outFile, 'Invalid Input: RDBW ', Pc - 2 : 1);
					flag := false;
					StopProgram := true
					END; {else}
				i := i - 1
				END; {while}
			IF flag THEN
				IF addr >= 0 THEN
					wordtobytes(wd, Mem[addr], Mem[addr + 1])
				ELSE
					Regs[reg] := wd
			END;
		4,                              {RDOD--read octal digit}
		5:                              {RDOW--read octal word}
			BEGIN
			writeln(outFile, 'Operations RDOD and RDOW not presently', 'implemented, at ', Pc - 2 : 1);
			StopProgram := true
			END;
		6:  BEGIN			{RDHD--read hex digit}
			read(inFile, ch);
			IF ch IN ['A'..'F', 'a'..'f', '0'..'9'] THEN BEGIN
				CASE ch OF
				'A', 'a': 
					wd := [15, 13];
				'B', 'b': 
					wd := [15, 13, 12];
				'C', 'c': 
					wd := [15, 14];
				'D', 'd': 
					wd := [15, 14, 12];
				'E', 'e': 
					wd := [15, 14, 13];
				'F', 'f': 
					wd := [15, 14, 13, 12];
				'0': 
					wd := [];
				'1': 
					wd := [12];
				'2': 
					wd := [13];
				'3': 
					wd := [13, 12];
				'4': 
					wd := [14];
				'5': 
					wd := [14, 12];
				'6': 
					wd := [14, 13];
				'7': 
					wd := [14, 13, 12];
				'8': 
					wd := [15];
				'9': 
					wd := [15, 12]
				END; {case of ch}
				IF addr >= 0 THEN BEGIN
				{Mem[addr] := Mem[addr] - [15, 14, 13, 12];}
				{Mem[addr] := Mem[addr] + wd}
					END
				ELSE BEGIN
					Regs[reg] := Regs[reg] - [15, 14, 13, 12];
					Regs[reg] := Regs[reg] + wd
					END {else}
				END
			ELSE BEGIN
				writeln(outFile, 'Invalid Input: RDHD ', Pc - 2 : 1);
				StopProgram := true
				END {else}
			END;
		7:  BEGIN			{RDHW--read hex word}
			i := 15;
			flag := true;
			wd := [];
			WHILE (i >= 1) AND flag DO BEGIN
				read(inFile, ch);
				IF ch IN ['A'..'F'] THEN
					ival := ord(ch) - ord('A') + 10
				ELSE IF ch IN ['a'..'f'] THEN
					ival := ord(ch) - ord('a') + 10
				ELSE IF ch IN ['0'..'9'] THEN
					ival := ord(ch) - ord('0')
				ELSE BEGIN
					writeln(outFile, 'Invalid Input: RDHW ', Pc - 2 : 1);
					flag := false;
					StopProgram := true
					END; {else}
				IF flag THEN BEGIN
					IF ival > 7 THEN BEGIN
						wd := wd + [i];
						ival := ival - 7
						END;
					i := i - 1;
					IF ival > 3 THEN BEGIN
						wd := wd + [i];
						ival := ival - 3
						END;
					i := i - 1;
					IF ival > 1 THEN BEGIN
						wd := wd + [i];
						ival := ival - 1
						END;
					i := i - 1;
					IF ival > 0 THEN
						wd := wd + [i];
					i := i - 1
					END {if}
				END; {while}
			IF addr >= 0 THEN
				wordtobytes(wd, Mem[addr], Mem[addr + 1])
			ELSE
				Regs[reg] := wd
			END;
		8:  BEGIN			{RDCH--read ASCII character}
			IF addr >= 0 THEN
				bytestoword(Mem[addr], Mem[addr + 1], Ccr.wf)
			ELSE
				Ccr.wf := Regs[reg];
			read(inFile, Ccr.c[1]);
			IF addr >= 0 THEN
				wordtobytes(Ccr.wf, Mem[addr], Mem[addr + 1])
			ELSE
				Regs[reg] := Ccr.wf
			END;
		9:  BEGIN			{RDST--read ASCII String}
			IF addr < 0 THEN BEGIN
				writeln(outFile, 'Invalid Address: RDST ', Pc - 2 : 1);
				StopProgram := true
				END
			ELSE BEGIN
				flag := true;
				WHILE NOT eoln DO
					IF flag THEN BEGIN
						flag := false;
						read(inFile, Ccr.c[1])
						END
					ELSE BEGIN
						flag := true;
						read(inFile, Ccr.c[2]);
						wordtobytes(Ccr.wf, Mem[addr], Mem[addr + 1]);
						addr := (addr + 2) MOD MEMSIZ
						END; {else}
				IF flag THEN
					Ccr.c[1] := chr(0)
				ELSE
					Ccr.c[2] := chr(0);
				wordtobytes(Ccr.wf, Mem[addr], Mem[addr + 1])
				END; {else}
			readln(inFile);
			END;
		10:                           {RDIN--read instruction}
			BEGIN
			writeln(outFile, 'Operation RDIN not presently implemented', ', at ', Pc - 2 : 1);
			StopProgram := true
			END;
		11:                           {RDNL--read newline}
			readln(inFile);
		END {case of rmode}
		END
	ELSE BEGIN
		writeln(outFile, 'Invalid Read Mode Detected at ', Pc - 2 : 1);
		StopProgram := true
		END
	END; { rdop }


	PROCEDURE wrop (wmode, amode, reg: integer; VAR w2: integer);
    {perform a write operation}
		VAR
			addr: integer;
			ival, bval, i: integer;
			wd: word;
			ch: char;
	BEGIN
	IF wmode IN [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11] THEN BEGIN
		addr := eac(amode, reg, w2);
		CASE wmode OF
		0:  BEGIN			{WRI}
			IF addr >= 0 THEN
				bytestoword(Mem[addr], Mem[addr + 1], Icr.wf)
			ELSE
				Icr.wf := Regs[reg];
			write(outFile, Icr.sintf : 1)
			END;
		1:  BEGIN			{WRF}
			IF addr >= 0 THEN BEGIN
{    Rcr.b2 := [];}
{    Rcr.b3 := [];}
				Rcr.b1 := Mem[addr];
				Rcr.b4 := Mem[addr + 1];
				Rcr.b5 := Mem[addr + 2];
				Rcr.b6 := Mem[addr + 3];
				END
			ELSE BEGIN
				wordtobytes(Regs[reg], Rcr.b1, Rcr.b4);
				wordtobytes(Regs[(reg + 1) MOD REGSIZ], Rcr.b5, Rcr.b6);
				END; {if}
			convertWtoExternal(Rcr);
			write(outFile, Rcr.rf)
			END;
		2:  BEGIN			{WRBD}
			IF addr >= 0 THEN
				bytestoword(Mem[addr], Mem[addr + 1], wd)
			ELSE
				wd := Regs[reg];
			IF 15 IN wd THEN
				write(outFile, '1')
			ELSE
				write(outFile, '0')
			END;
		3:  BEGIN			{WRBW}
			IF addr >= 0 THEN
				bytestoword(Mem[addr], Mem[addr + 1], wd)
			ELSE
				wd := Regs[reg];
			FOR i := 0 TO 15 DO
				IF i IN wd THEN
					write(outFile, '1')
				ELSE
					write(outFile, '0')
			END;
		4,                              {WROD--write octal digit}
		5:                              {WROW--write octal word}
			BEGIN
			writeln(outFile, 'Operations WROD and WROW not presently', 'implemented, at ', Pc - 2 : 1);
			StopProgram := true
			END;
		6:  BEGIN			{WRHD}
			IF addr >= 0 THEN
				bytestoword(Mem[addr], Mem[addr + 1], wd)
			ELSE
				wd := Regs[reg];
			bval := 8;
			ival := 0;
			FOR i := 15 DOWNTO 12 DO BEGIN
				IF i IN wd THEN
					ival := ival + bval;
				bval := bval DIV 2
				END; {for}
			write(outFile, ival : 1) {hex}
			END;
		7:  BEGIN			{WRHW}
			IF addr >= 0 THEN
				bytestoword(Mem[addr], Mem[addr + 1], wd)
			ELSE
				wd := Regs[reg];
			bval := maxSmallInt;
			ival := 0;
			FOR i := 15 DOWNTO 0 DO BEGIN
				IF i IN wd THEN
					ival := ival + bval;
				bval := bval DIV 2
				END; {for}
			IF ival < 4096 THEN
				write(outFile, '0');
			IF ival < 256 THEN
				write(outFile, '0');
			IF ival < 16 THEN
				write(outFile, '0');
			write(outFile, ival : 1) {hex}
			END;
		8:				{WRCH--write character}
			BEGIN
			IF addr >= 0 THEN
				bytestoword(Mem[addr], Mem[addr + 1], Ccr.wf)
			ELSE
				Ccr.wf := Regs[reg];
			write(outFile, Ccr.c[1])
			END;
		9:				{WRST--write string}
			BEGIN
			ch := ' ';
			WHILE ord(ch) <> 0 DO BEGIN
				IF addr >= 0 THEN BEGIN
					bytestoword(Mem[addr], Mem[addr + 1], Ccr.wf);
					addr := (addr + 2) MOD MEMSIZ
					END
				ELSE BEGIN
					Ccr.wf := Regs[reg];
					reg := (reg + 1) MOD REGSIZ
					END; {if}
				IF ord(Ccr.c[1]) <> 0 THEN BEGIN
					write(outFile, Ccr.c[1]);
					IF ord(Ccr.c[2]) <> 0 THEN
						write(outFile, Ccr.c[2])
					ELSE
						ch := chr(0)
					END
				ELSE
					ch := chr(0)
				END {while}
			END;
		11:				{WRNL--write newline}
			writeln(outFile)
		END {case of wmode}
		END
	ELSE BEGIN
		writeln(outFile, 'Invalid Write Mode Dectected at ', Pc - 1 : 1);
		StopProgram := true
		END
	END; { wrop }


	PROCEDURE execute;
		VAR
			opcode: integer;
			f1: integer;
			f2: integer;
			f3: integer;
			w2: integer;
	BEGIN
	opcode := 0;
	IF 15 IN Ir THEN
		opcode := 16;
	IF 14 IN Ir THEN
		opcode := opcode + 8;
	IF 13 IN Ir THEN
		opcode := opcode + 4;
	IF 12 IN Ir THEN
		opcode := opcode + 2;
	IF 11 IN Ir THEN
		opcode := opcode + 1;
	f1 := 0;
	IF 10 IN Ir THEN
		f1 := 8;
	IF 9 IN Ir THEN
		f1 := f1 + 4;
	IF 8 IN Ir THEN
		f1 := f1 + 2;
	IF 7 IN Ir THEN
		f1 := f1 + 1;
	f2 := 0;
	IF 6 IN Ir THEN
		f2 := 4;
	IF 5 IN Ir THEN
		f2 := f2 + 2;
	IF 4 IN Ir THEN
		f2 := f2 + 1;
	f3 := 0;
	IF 3 IN Ir THEN
		f3 := 8;
	IF 2 IN Ir THEN
		f3 := f3 + 4;
	IF 1 IN Ir THEN
		f3 := f3 + 2;
	IF 0 IN Ir THEN
		f3 := f3 + 1;
	w2 := 0;
	CASE opcode OF
	INDENT, IA, IS, IM, ID, FN, FA, FS, FM, FD, BI, BO, BA, IC, FC, JSR, BKT, LD, STO, LDA, FIX, FLT, TRNG: 
		twoaddr(opcode, f1, f2, f3, w2);
	J: 
		jop(f1, f2, f3, w2);
	SR: 
		srop(f1, f2, f3);
	SL: 
		slop(f1, f2, f3);
	RD: 
		rdop(f1, f2, f3, w2);
	WR: 
		wrop(f1, f2, f3, w2);
	Halt: 
		StopProgram := true;
	28, 29, 30:  BEGIN
		writeln(outFile, 'Bad opcode', Pc - 2 : 1);
		StopProgram := true
		END
	END; {case}
	IF Trace THEN BEGIN
		writeln(outFile, '----------------------------------------');
		writeinst(oldPc, opcode, f1, f2, f3, w2);
		writeln(outFile, '----------------------------------------');
		displayregs
		END {for}
	END; {procedure execute}

	PROCEDURE rdmem;
		VAR
			memf: FILE OF byterange;

			i: integer;
	BEGIN
    {assign ( memf, 'OBJ' );}
	reset(memf, 'OBJ');
	IF eof(memf) THEN
		MemImage := false;
	i := 0;
	WHILE NOT eof(memf) DO BEGIN
		read(memf, Bcr.hintf);
		Mem[i] := Bcr.bytef;
		i := i + 1
		END {while}
	END; {procedure rdmem}



BEGIN {main program}
showDrawing;
moveTo(40, 20);
drawstring('MACC2 Simulator');
moveTo(40, 35);
DrawString('Name of the Input Data File?');
MemoryFileName := oldFileName('Name of the Input File? ');
IF MemoryFileName <> '' THEN BEGIN
	Reset(inFile, MemoryFileName);
	END;
clearLine(35);
moveTo(40, 35);
DrawString('Name of the Output Text File? ');
ListingFileName := NewFileName('Name of the Output Text File? ');
IF ListingFileName = '' THEN
	Rewrite(outFile, 'Output')
ELSE
	Rewrite(outFile, ListingFileName);
moveTo(220, 20);
drawString('WORKING...');
hideAll;
showText;

MemImage := true;
Dump := false;
Trace := false;
rdmem;
Lt := false;
Eq := false;
Gt := false;
Pc := 0;
StopProgram := false;
IF MemImage THEN BEGIN
	FOR RegIndex := 0 TO 15 DO
		Regs[RegIndex] := [];
	REPEAT
		oldPc := Pc;
		fetchword(Ir);         {fetch the next instruction}
		execute;
{   IF Trace THEN BEGIN}
{    writeln;}
{    writeln('Press <ENTER> to contine.');}
{    readln;}
{    writeln}
{    END}
	UNTIL StopProgram    {execute the current inst}
	END;
IF Dump THEN BEGIN
       {assign ( Dumpf, 'DUMPFILE' );}
	rewrite(Dumpf, 'DUMPFILE');
	FOR i := 1 TO MEMSIZ - 1 DO BEGIN
		Bcr.bytef := Mem[i];
		write(Dumpf, Bcr.hintf)
		END;{for}
	close(dumpf);
	END;{if}
close(outfile);
noError := GetFInfo(ListingFileName, vRefNum, finderInfo);
finderInfo.fdType := 'TEXT';
finderInfo.fdCreator := 'QED1';
noError := SetFInfo(ListingFileName, vRefNum, finderInfo);
END.